/*
 * Replacement Debug Interface

 * Copyright (C) 2022-2024 Yeoldetoast

 * This file is part of YeoldeDink

 * Yeoldedink is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.

 * Yeoldedink is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <vector>

#include "SDL.h"
//#include "curl/curl.h"
#ifdef HAVE_SDL_GPU
#include "SDL_gpu.h"
#endif
#ifdef HAVE_MIXER_EXT
#include <SDL2/SDL_mixer_ext.h>
#else
#include <SDL_mixer.h>
#endif

#include <cereal/cereal.hpp>
#include <cereal/types/unordered_map.hpp>
#include <cereal/types/memory.hpp>
#include <cereal/archives/binary.hpp>
#include <cereal/archives/json.hpp>
#include "soloud/soloud.h"
#include "soloud/soloud_wav.h"
#include "soloud/soloud_freeverbfilter.h"

//Engine includes
#include "debug_imgui.h"
#include "game_state.h"
#include "game_engine.h"
#include "live_sprites_manager.h"
#include "input.h"
#include "log.h"
#include "dinkc.h"
#include "dinkc_console.h"
#include "scripting.h"
#include "freedink.h"
#include "gfx.h"
#include "gfx_sprites.h"
#include "gfx_fonts.h"
#include "update_frame.h"
#include "live_screen.h"
#include "inventory.h"
#include "ImageLoader.h"
#include "status.h"
#include "sfx.h"
#include "bgm.h"
#include "DMod.h"
#include "savegame.h"
#include "paths.h"
#include "app.h"
#include "editor_screen.h"
#include "gfx_palette.h"
#include "hardness_tiles.h"
#include "dinklua.h"
#include "dinklua_bindings.h"
#include "scripting.h"
#include "script_bindings.h"

//header-only libraries
#include "imgui.h"
#include "imgui_impl_sdl2.h"
#include "imgui_internal.h"
#include "implot.h"
#include "SimpleIni.h"
#include "color/color.hpp"
#include "random.hpp"
#include "spdlog/spdlog.h"
#include <glob.hpp>

//scripting
#include <sol/sol.hpp>

using Random = effolkronium::random_static;

//Settings struct for serialising
struct debugsettings dbg;

//For the log window
ImGuiTextBuffer Buf, errBuf, dbgBuf, infBuf;;
ImGuiTextFilter Filter;
ImVector <int> LineOffsets;
bool AutoScroll = true;
//For transparent overlay and console
char debugbuf[128];
static int DinkcHistoryPos = -1;
ImVector<char*> DinkcHistory;
char luabuf[128];
static int LuaHistoryPos = -1;
ImVector<char*> LuaHistory;
static int ts = 1;
char sfbuf[128];
//For various other window settings
bool render_tiles = true;
bool drawhard;
bool globinfo;

//For music and other things
static int modorder;
static float sfxvol;
static double muspos;
static int zero = 0;
static int tilemax = 5248;
static int htilemax = 800;
Mix_MusicType tracker = MUS_MOD;
char midichlorian[40];
int midiplayer;
static bool jukesel = false;
static int juker;
ImVector<char[400]> jlist;

//Speech synthesis
int spwave;
int spfreq;
int spdec;
int spspeed;

//for the var/sprite editors
int varedit = 1;
int spriedit = 2;
int sprigrid = 0;
int mspried = 1;
bool spoutline;
bool sprilivechk;
bool hboxprev, ddotprev;
//copy paste buffers
editor_sprite spritebuf[100 + 1];
editor_screen screenbuf;
static int xmax = 620;
static int ymax = 400;
static int ypos;
static ImPlotRect hbox, trimbox;
static bool sprplotedit;
static bool mspriunused;
//Tab item flags
ImGuiTabItemFlags_ flag = ImGuiTabItemFlags_None;
ImGuiTabItemFlags_ mflag = ImGuiTabItemFlags_None;

//Script ed
static int myscript;
struct mb_interpreter_t* bas = NULL;
char mbconsole[200];

//For dink.ini viewer
bool freeslots;

//For gamepad
int lbutton = ACTION_BUTTON7;
const char* controllercmds[] = { "Talk", "Magic", "Inventory", "Menu", "Map", "button7.c", "button8.c", "button9.c", "button10.c" };

//For sprite overrider
static bool sprovermod;
static int sproversel = 1;

//For colour palette thingo
static int palcol;
static int mycol;
static bool autopal;
ImVec4 mypalcol;
SDL_Color mypal[256] = {};
char bmppal[100];

//For the box previewer
rect boxpreview;
ImPlotRect bprev;
static bool bprevview;

//For the metrics, crap temp vars too
//static float t = 0;
//static float shistory = 60.0f;
const char* dispfont;
const char* debugfont;
char debugfon[200];
int debug_fontsize;
int displayx;
int displayy;
bool dispfilt;
bool dispstretch;
int audiosamp;
int audiobuf;
int audiovol;
int sfx_vol;
bool dbgtxtwrite;
bool dbgconpause;
bool dbginfowindow;
float pan, muspeed, muhandle;

//For hard.dat and tile viewer
static int htile;
static int tindexscr = 1;
editor_screen_tilerefs mytiles;

//Toggleable settings
bool debug_meminfo;
bool debug_pinksquares;
bool debug_drawsprites = true;
bool debug_drawgame = true;
bool debug_gamewindow;
bool debug_paused;
bool debug_conpause;
bool debug_infhp;
bool debug_magcharge;
bool debug_invspri;
bool debug_noclip;
bool debug_vision;
bool debug_screenlock;
bool debug_assratio;
bool debug_walkoff;
bool debug_audiopos;
bool debug_speechsynth;
bool debug_returnint;

//Toggles for smaller windows
bool ultcheat;
bool graphicsopt;
bool dswitch;
bool logwindow;
bool keeb;
bool scripfo;
bool sprinfo;
bool variableinfo;
bool sfxinfo;
bool metrics;
bool mapsquares;
bool maptiles;
bool sysinfo;
bool mapsprites;
bool dinkini;
bool helpbox;
bool aboutbox;
bool sproverride;
bool paledit;
bool iniconf;
static bool boxprev;
static bool scratchpad;
static bool htiles;
static bool shaderwindow;
static bool jukebox;
static bool magicitems;
static bool luaviewer;
static bool spreditors;

//TODO: something
#define rinfo(script) ((struct refinfo*)(sinfo[script]->data))

//SDL_Texture* tilescreen[128] = NULL;
//SDL_Texture* map_tiles[96] = NULL;

//Have to save sprites, scrip, items, SFX, music probably, variables, editor stuff, map screen, dinkc?
/* void savestate() {
	std::ofstream os(paths_dmodfile("autosave.dink"), std::ios::binary);
	cereal::BinaryOutputArchive archive(os);
	 archive(play.spmap);
	 archive(play.var);
	 //archive(spr);
}

void loadstate() {
	try {
	std::ifstream is(paths_dmodfile("autosave.dink"), std::ios::binary);
	cereal::BinaryInputArchive archive(is);
	archive(play.spmap);
	archive(play.var);
	}
	catch (const std::exception & e) {
		log_error("Could not load save state!");
	}
} */

//The most important feature
void play_sneedle() {
	playsfx(paths_pkgdatafile("sneedle.mp3"), 0, 0);
}

int LuaConsoleCallback(ImGuiInputTextCallbackData *data)
{
	const int prev_history_pos = LuaHistoryPos;
	if (data->EventFlag == ImGuiInputTextFlags_CallbackHistory) {
	if (data->EventKey == ImGuiKey_UpArrow)
	{
		if (LuaHistoryPos == -1)
			LuaHistoryPos = LuaHistory.Size - 1;
		else if (LuaHistoryPos > 0)
			LuaHistoryPos--;
	}
	else if (data->EventKey == ImGuiKey_DownArrow)
	{
		if (LuaHistoryPos != -1)
			if (++LuaHistoryPos >= LuaHistory.Size)
				LuaHistoryPos = -1;
	}

	// A better implementation would preserve the data on the current input line along with cursor position.
	if (prev_history_pos != LuaHistoryPos)
	{
		const char *history_str = (LuaHistoryPos >= 0) ? LuaHistory[LuaHistoryPos] : "";
		data->DeleteChars(0, data->BufTextLen);
		data->InsertChars(0, history_str);
	}
	}
}

int DinkcConsoleCallback(ImGuiInputTextCallbackData *data)
{
	const int prev_history_pos = DinkcHistoryPos;
	if (data->EventFlag == ImGuiInputTextFlags_CallbackHistory) {
	if (data->EventKey == ImGuiKey_UpArrow)
	{
		if (DinkcHistoryPos == -1)
			DinkcHistoryPos = DinkcHistory.Size - 1;
		else if (DinkcHistoryPos > 0)
			DinkcHistoryPos--;
	}
	else if (data->EventKey == ImGuiKey_DownArrow)
	{
		if (DinkcHistoryPos != -1)
			if (++DinkcHistoryPos >= DinkcHistory.Size)
				DinkcHistoryPos = -1;
	}

	// A better implementation would preserve the data on the current input line along with cursor position.
	if (prev_history_pos != DinkcHistoryPos)
	{
		const char *history_str = (DinkcHistoryPos >= 0) ? DinkcHistory[DinkcHistoryPos] : "";
		data->DeleteChars(0, data->BufTextLen);
		data->InsertChars(0, history_str);
	}
	}
}

void refresh_map_tiles() {
	for (int i = 0; i < 128; i++) {
		//tilescreen[i * ts - 1] = SDL_CreateTextureFromSurface(SDL_GetRenderer(g_display->window), g_dmod.bgTilesets.slots[i * ts - 1]->screenshot());
	}
}

void gameload(int slot) {
	scripting_kill_all_scripts_for_real();
	kill_repeat_sounds_all();
	load_game(slot);
	*pupdate_status = 1;
	draw_status_all();
}
//stolen from curl
static size_t write_data(void *ptr, size_t size, size_t nmemb, void *stream)
{
  size_t written = fwrite(ptr, size, nmemb, (FILE *)stream);
  return written;
}

//For jukebox
void jukelist() {
	for (auto& p : glob::glob({paths_dmodfile("SOUND/*.mid"), paths_dmodfile("SOUND/*.MID"), paths_dmodfile("SOUND/*.spc"), paths_dmodfile("SOUND/*.mod"), paths_dmodfile("SOUND/*.xm"), paths_dmodfile("SOUND/*.it"), paths_dmodfile("SOUND/*.s3m"), paths_dmodfile("SOUND/*.ogg"), paths_dmodfile("SOUND/*.opus"), paths_dmodfile("SOUND/*.mp3")})) {
		char buf[400];
		#ifndef _WIN32
		sprintf(buf, p.filename().c_str());
		#endif
		jlist.push_back(buf);
		//spdlog::info("Jukebox: {}", jlist.back());
	}
}

//Alter HSV for entire custom palette
void changehsv(int param, int amount, int space)
{
  // 0 is h, 1 s, 2 v/i/l
  for (int c = 0; c < 256; c++)
  {
	  // Convert colour to HSV
	  color::rgb<std::uint8_t> c1({mypal[c].r, mypal[c].g, mypal[c].b});
	  color::hsv<std::uint8_t> c2;
	  // Or HSI/HSL
	  if (space == 1)
	  color::hsl<std::uint8_t> c2;
	  if (space == 2)
	  color::hsi<std::uint8_t> c2;

	  c2 = c1;
	  // Change its parameter
	  c2[param] += amount;
	  // Convert back to RGB
	  c1 = c2;
	  mypal[c].r = ::color::get::red(c1);
	  mypal[c].g = ::color::get::green(c1);
	  mypal[c].b = ::color::get::blue(c1);
  }
}

void livesprtomap(int srcspr) {
	//Saves a live sprite to an editor sprite. Not all values are transferred of course
	cur_ed_screen.sprite[spr[srcspr].sp_index].brain = spr[srcspr].brain;
	cur_ed_screen.sprite[spr[srcspr].sp_index].x = spr[srcspr].x;
	cur_ed_screen.sprite[spr[srcspr].sp_index].y = spr[srcspr].y;
	cur_ed_screen.sprite[spr[srcspr].sp_index].size = spr[srcspr].size;
	cur_ed_screen.sprite[spr[srcspr].sp_index].seq = spr[srcspr].pseq;
	cur_ed_screen.sprite[spr[srcspr].sp_index].frame = spr[srcspr].pframe;
	cur_ed_screen.sprite[spr[srcspr].sp_index].touch_damage = spr[srcspr].touch_damage;
	cur_ed_screen.sprite[spr[srcspr].sp_index].gold = spr[srcspr].gold;
	cur_ed_screen.sprite[spr[srcspr].sp_index].exp = spr[srcspr].exp;
	cur_ed_screen.sprite[spr[srcspr].sp_index].que = spr[srcspr].que;
	cur_ed_screen.sprite[spr[srcspr].sp_index].hitpoints = spr[srcspr].hitpoints;
	cur_ed_screen.sprite[spr[srcspr].sp_index].sound = spr[srcspr].sound;
	//trimming
	cur_ed_screen.sprite[spr[srcspr].sp_index].alt = spr[srcspr].alt;

}

//From imgui
static void HelpMarker(const char* desc)
{
    ImGui::TextDisabled("(?)");
    if (ImGui::IsItemHovered(ImGuiHoveredFlags_DelayShort) && ImGui::BeginTooltip())
    {
        ImGui::PushTextWrapPos(ImGui::GetFontSize() * 35.0f);
        ImGui::TextUnformatted(desc);
        ImGui::PopTextWrapPos();
        ImGui::EndTooltip();
    }
}

//Stolen from the editor. Writes to dink.ini
static void add_text() {
	char death[150];
	char filename[10];
	//I should have just used getpic()
	sprintf(death, "SET_SPRITE_INFO %d %d %d %d %d %d %d %d\n",
		spr[spriedit].pseq, spr[spriedit].pframe,
		k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].xoffset,
		k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].yoffset,
		k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.left,
		k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.top,
		k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.right,
		k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.bottom);

	strcpy(filename, "dink.ini");

	FILE* fp = paths_dmodfile_fopen(filename, "ab");
	if (fp != NULL) {
		fwrite(death, strlen(death), 1, fp);
		fclose(fp);
	} else {
		perror("add_text");
	}
}

//also taken from the editor
int add_new_map() {
	int loc_max = 0;
	for (int i = 0; i < 768 + 1; i++)
		if (g_dmod.map.loc[i] > loc_max)
			loc_max = g_dmod.map.loc[i];

	int loc_new = loc_max + 1;
	if (loc_new > 768)
		return -1;

	save_screen(g_dmod.map.map_dat.c_str(), loc_new);
	return loc_new;
}

//Runs every frame in the render loop to show the gui stuff
void dbg_imgui(void) {
    //Only show stuff if debug is on and the game has loaded to title screen
	if (debug_mode && mode) {
//Show the mouse
ImGuiIO& io = ImGui::GetIO();
io.MouseDrawCursor = true;
//Main Menu bar
ImGui::BeginMainMenuBar();
if (ImGui::BeginMenu("File"))
{
    if (ImGui::MenuItem("Hide menubar and debug interface", "Alt+D"))
    log_debug_off();

	ImGui::MenuItem("Pause", NULL, &debug_paused);

	if (ImGui::MenuItem("Restart"))
		game_restart();
	if (ImGui::MenuItem("Quit", "Alt+Q")) {
	    SDL_Event ev;
	    ev.type = SDL_QUIT;
	    SDL_PushEvent(&ev);
	}

	ImGui::EndMenu();
}

if (ImGui::BeginMenu("Debug"))
{
	ImGui::MenuItem("Developer mode", NULL, &dbg.devmode);
	ImGui::Separator();
	ImGui::MenuItem("Info overlay", NULL, &dbg.dswitch);
	ImGui::MenuItem("Debug Log", NULL, &logwindow);
	if (dbg.devmode)
	{
		ImGui::MenuItem("Game popout window", NULL, &debug_gamewindow);
		ImGui::MenuItem("Ultimate Cheat", NULL, &ultcheat);
		ImGui::Separator();
		ImGui::MenuItem("Pause while console activated", NULL, &dbg.debug_conpause);
		if (ImGui::BeginMenu("Log Level"))
		{
			if (ImGui::MenuItem("Errors only"))
				log_debug_on(0);
			if (ImGui::MenuItem("Standard"))
				log_debug_on(1);
			if (ImGui::MenuItem("Verbose"))
				log_debug_on(2);

			ImGui::EndMenu();
		}
		if (ImGui::BeginMenu("Visual options"))
		{
			ImGui::MenuItem("Render game in main window", NULL, &debug_drawgame);
			ImGui::MenuItem("Pink squares", NULL, &dbg.debug_pinksquares);
			ImGui::MenuItem("Hard.dat overlay", "Alt+H", &drawhard);
			ImGui::MenuItem("Alternate text display", NULL, &dbg.alttext);
			ImGui::EndMenu();
		}
		if (ImGui::BeginMenu("Sprite rendering"))
		{
			// ImGui::MenuItem("Render tiles and background", NULL, &render_tiles);
			if (ImGui::MenuItem("Draw invisible sprites", NULL, &debug_invspri))
			{
				game_place_sprites_background();
				draw_screen_game_background();
			}
			ImGui::MenuItem("Ignore sprite visions ", NULL, &debug_vision);
			ImGui::EndMenu();
		}
		// ImGui::MenuItem("Write to debug.txt", NULL, &dbg.debug_conwrite);
	}
	ImGui::EndMenu();
}

if (ImGui::BeginMenu("Display")) {
	ImGui::MenuItem("Texture filtering", NULL, &dbg.debug_interp);
	//ImGui::MenuItem("Shaders", NULL, &shaderwindow);
	ImGui::MenuItem("Ignore aspect ratio", NULL, &dbg.debug_assratio);
	ImGui::MenuItem("Screen transition", NULL, &dbg.screentrans);
	if (!truecolor)
	ImGui::BeginDisabled();
	ImGui::MenuItem("'Smooth' Shadows", NULL, &dbg.debug_goodshadows);
	ImGui::MenuItem("Smooth fade down/up", NULL, &dbg.altfade);
	if (!truecolor)
	ImGui::EndDisabled();
	//ImGui::MenuItem("Warp fade", NULL, &debug_warpfade);
	ImGui::SeparatorText("Windowed Resolution");
	if (ImGui::MenuItem("640x480")) {
		SDL_SetWindowFullscreen(g_display->window, 0);
		SDL_SetWindowSize(g_display->window, 640, 480);
		SDL_SetWindowPosition(g_display->window, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
	}

	if (ImGui::MenuItem("1280x960")) {
	SDL_SetWindowFullscreen(g_display->window, 0);
	SDL_SetWindowSize(g_display->window, 1280, 960);
	SDL_SetWindowPosition(g_display->window, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
	}
	if (ImGui::MenuItem("1920x1440")) {
		SDL_SetWindowFullscreen(g_display->window, 0);
		SDL_SetWindowSize(g_display->window, 1920, 1440);
		SDL_SetWindowPosition(g_display->window, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED);
	}
	ImGui::Separator();
	if (ImGui::MenuItem("Toggle full screen display")) {
		g_display->toggleFullScreen();
	}
	if (ImGui::MenuItem("Make window borderless")) {
		SDL_SetWindowBordered(g_display->window, SDL_FALSE);
	}

	ImGui::EndMenu();
}

if (ImGui::BeginMenu("System"))
{
	//ImGui::MenuItem("Speech synthesis", NULL, &debug_speechsynth);
	//ImGui::Separator();
    //ImGui::MenuItem("Debug Mode Options", NULL, &dswitch);
	//ImGui::MenuItem("Graphics Options", NULL, &graphicsopt);
    ImGui::MenuItem("Gamepad and Joystick", NULL, &keeb);
    ImGui::MenuItem("Audio", NULL, &sfxinfo);
	#ifndef _WIN32
	ImGui::MenuItem("Jukebox", NULL, &jukebox);
	#endif
	ImGui::Separator();
	ImGui::MenuItem("System Information", NULL, &sysinfo);

	ImGui::EndMenu();
}


if (dbg.devmode) {
if (ImGui::BeginMenu("Engine Data"))
{

	ImGui::MenuItem("Sprite List", NULL, &sprinfo);
	ImGui::MenuItem("Sprite Editor", NULL, &spreditors);
	ImGui::MenuItem("Variables", NULL, &variableinfo);
	ImGui::MenuItem("Scripts", NULL, &scripfo);
	//if (!dinklua_enabled)
	//ImGui::BeginDisabled();
	//ImGui::MenuItem("Lua state", NULL, &luaviewer);
	//if (!dinklua_enabled)
	//ImGui::EndDisabled();

	if (mode < 3)
	ImGui::BeginDisabled();
	ImGui::MenuItem("Editor Overrides", NULL, &sproverride);
	ImGui::MenuItem("Tiles", NULL, &maptiles);
	ImGui::MenuItem("Magic and Items", NULL, &magicitems);
    ImGui::Separator();
	if (mode < 3)
	ImGui::EndDisabled();
	ImGui::MenuItem("Map Index", NULL, &mapsquares);
	ImGui::MenuItem("Hard.dat", NULL, &htiles);
	ImGui::MenuItem("Dink.ini", NULL, &dinkini);
	ImGui::MenuItem("Colour Palette", NULL, &paledit);

	ImGui::EndMenu();
}
}

//if (ImGui::BeginMenu("Toolbox")) {
//	ImGui::MenuItem("Box previewer", NULL, &boxprev);
	//ImGui::MenuItem("Script scratchpad", NULL, NULL);

//	ImGui::EndMenu();
//}
if (dbg.devmode) {
if (ImGui::BeginMenu("Speed"))
{
	if (ImGui::MenuItem("Normal Speed"))
		game_set_normal_speed();
	if (ImGui::MenuItem("High speed"))
		game_set_high_speed();
	if (ImGui::MenuItem("Slow speed"))
		game_set_slow_speed();
	ImGui::Separator();
	ImGui::MenuItem("Remove framerate limiter", NULL, &dbg.framelimit);
	ImGui::MenuItem("Use old sprite ranking", NULL, &dbg.ranksprites);
	ImGui::EndMenu();
}
}

// Only show if game has started. Does weird stuff otherwise and I can't be bothered to fix it
if (ImGui::BeginMenu("Quick Save/Load"))
{
	if (ImGui::MenuItem("Autosave to slot 10 on screen change", NULL, &dbg.autosave)) {
		//autosave only works with screen transitions on
		dbg.screentrans = true;
	}
	ImGui::Separator();

	if (mode < 2)
		ImGui::BeginDisabled();
	if (ImGui::MenuItem("QuickSave #1"))
		save_game(1337);
	if (ImGui::MenuItem("QuickSave #2"))
		save_game(1338);
	if (ImGui::MenuItem("QuickSave #3"))
		save_game(1339);
	ImGui::Separator();
	if (ImGui::MenuItem("Load #1"))
	{
		gameload(1337);
	}

	if (ImGui::MenuItem("Load #2"))
	{
		gameload(1338);
	}

	if (ImGui::MenuItem("Load #3"))
	{
		gameload(1339);
	}

	if (mode < 2)
		ImGui::EndDisabled();

	ImGui::EndMenu();
}

if (ImGui::BeginMenu("Help"))
{
ImGui::MenuItem("Contents", NULL, &helpbox);
ImGui::MenuItem("About", NULL, &aboutbox);
ImGui::EndMenu();
}

if (console_active == 1) {
	if (ImGui::BeginMenu("Warning! Console Active!")) {
		if (ImGui::MenuItem("Deactivate Console", "Alt+C"))
			dinkc_console_hide();
		ImGui::EndMenu();
	}

}


ImGui::EndMainMenuBar();

//ImGui::ShowDemoWindow();
//ImPlot::ShowDemoWindow();
//End of menus, start of windows and such
//Transparent overlay info window, stolen from the demo
if (dbg.dswitch) {
ImGuiWindowFlags window_flags = ImGuiWindowFlags_NoDecoration | ImGuiWindowFlags_AlwaysAutoResize | ImGuiWindowFlags_NoSavedSettings | ImGuiWindowFlags_NoFocusOnAppearing | ImGuiWindowFlags_NoNav;
if (dbg.corner != -1)
    {
        const float PAD = 10.0f;
        const ImGuiViewport* viewport = ImGui::GetMainViewport();
        ImVec2 work_pos = viewport->WorkPos; // Use work area to avoid menu-bar/task-bar, if any!
        ImVec2 work_size = viewport->WorkSize;
        ImVec2 window_pos, window_pos_pivot;
        window_pos.x = (dbg.corner & 1) ? (work_pos.x + work_size.x - PAD) : (work_pos.x + PAD);
        window_pos.y = (dbg.corner & 2) ? (work_pos.y + work_size.y - PAD) : (work_pos.y + PAD);
        window_pos_pivot.x = (dbg.corner & 1) ? 1.0f : 0.0f;
        window_pos_pivot.y = (dbg.corner & 2) ? 1.0f : 0.0f;
        ImGui::SetNextWindowPos(window_pos, ImGuiCond_Always, window_pos_pivot);
        window_flags |= ImGuiWindowFlags_NoMove;
    }
	ImGui::SetNextWindowBgAlpha(0.35f); // Transparent background
    if (ImGui::Begin("Example: Simple overlay", NULL, window_flags))
    {
        ImGui::Text("Debug mode enabled. Alt+D to disable, Alt+C for console");
        ImGui::Separator();
		//Change the colour according to framerate
		ImVec4 col = ImVec4(1.0f, 1.0f, 0.0f, 1.0f);
		float fram = ImGui::GetIO().Framerate;
		if (fram >= 50)
			col = ImVec4(0.1f, 1.0f, 0.0f, 1.0f);
		if (fram < 30)
			col = ImVec4(1.0f, 0.0f, 0.0f, 1.0f);
		ImGui::TextColored(col, "Application average %.3f ms/frame (%.1f FPS)", 1000.0f / ImGui::GetIO().Framerate, ImGui::GetIO().Framerate);
		ImGui::Text("Sprites: %d Moveman X%d X%d: Speed: %d Y%d Y%d, Screen: %d (%d)",
					last_sprite_created, spr[1].lpx[0],
					spr[1].lpy[0], spr[1].moveman, spr[1].lpx[1], spr[1].lpy[1], *pplayer_map, g_dmod.map.loc[*pplayer_map]);
		//time_t ct;
		//time(&ct);
		//ImGui::Text("%f", difftime(ct, time_start));
        //if (ImGui::IsMousePosValid()) {
		//	float x, y;
		//	SDL_RenderWindowToLogical(SDL_GetRenderer(g_display->window), io.MousePos.x, io.MousePos.y, &x, &y);
        //   ImGui::Text("Mouse Position: (%.1f,%.1f) (%d, %d)", io.MousePos.x, io.MousePos.y, x, y);
		//}
        //else
         //   ImGui::Text("Mouse Position: <invalid>");
        ImGui::Text("%s", log_getLastLog());
		ImGuiInputTextFlags input_text_flags = ImGuiInputTextFlags_EnterReturnsTrue | ImGuiInputTextFlags_CallbackHistory | ImGuiInputTextFlags_EscapeClearsAll;
		if (console_active) {
			if (debug_conpause)
				debug_paused = 1;
				if (dinkc_enabled) {
		if (ImGui::InputText("DinkC Console", debugbuf, IM_ARRAYSIZE(debugbuf), input_text_flags, &DinkcConsoleCallback))
			{
			ImGui::SetItemDefaultFocus();
			dinkc_execute_one_liner(debugbuf);
			DinkcHistoryPos = -1;
			DinkcHistory.push_back(strdup(luabuf));
			strcpy(debugbuf, "");
			}
			//Console off button
			ImGui::SameLine();
			if (ImGui::SmallButton("Off"))
				dinkc_console_hide();
				}
			if (dinklua_enabled) {
			if (ImGui::InputText("Lua Console", luabuf, IM_ARRAYSIZE(luabuf), input_text_flags, &LuaConsoleCallback)) {
				dinklua_execute_one_liner(luabuf);
				LuaHistoryPos = -1;
				LuaHistory.push_back(strdup(luabuf));
				strcpy(luabuf, "");
			}
			//Console hide button
			if (!dinkc_enabled) {
			ImGui::SameLine();
			if (ImGui::SmallButton("Off"))
				dinkc_console_hide();
				}
			}


		}
        if (ImGui::BeginPopupContextWindow())
        {
            if (ImGui::MenuItem("Custom",       NULL, dbg.corner == -1)) dbg.corner = -1;
            if (ImGui::MenuItem("Top-left",     NULL, dbg.corner == 0)) dbg.corner = 0;
            if (ImGui::MenuItem("Top-right",    NULL, dbg.corner == 1)) dbg.corner = 1;
            if (ImGui::MenuItem("Bottom-left",  NULL, dbg.corner == 2)) dbg.corner = 2;
            if (ImGui::MenuItem("Bottom-right", NULL, dbg.corner == 3)) dbg.corner = 3;
            if (ImGui::MenuItem("Close")) dbg.dswitch = false;
            ImGui::EndPopup();
        }
    }
    ImGui::End();
}
//Keyboard state and gamepad
if (keeb) {
ImGui::Begin("Gamepad and Joystick", &keeb);
if (ImGui::Button("Reinitialise input")) {
	joystick = 1;
	input_init();
}
//TODO: get rid of this
//int x;
//for (x = 0; x < 256; x++)
//	if (input_getcharstate(x))
//	ImGui::Text("Key %i", x);
	if (jinfo) {
			SDL_GameController* ginfo = SDL_GameControllerFromInstanceID(SDL_JoystickInstanceID(jinfo));
			ImGui::Text("Detected Name: %s", SDL_JoystickName(jinfo));
			ImGui::Text("Button map loaded: %s", SDL_GameControllerMapping(ginfo));
			if (SDL_JoystickHasRumble(jinfo)) {
				if (ImGui::Button("Test rumble")) {
				SDL_GameControllerRumble(ginfo, 0xFFFF, 0xFFFF, 500);
				SDL_GameControllerRumbleTriggers(ginfo, 0xFFFF, 0xFFFF, 900);
				}
				//ImGui::SameLine();
				ImGui::Checkbox("Rumble effect", &dbg.jrumble);
				}
			//ImGui::SameLine();
			if (ImGui::Button("Reload button map")) {
				int mapping = SDL_GameControllerAddMappingsFromFile(paths_pkgdatafile("gamecontrollerdb.txt"));
			}
			//if (ImGui::Button("Switch off gamepad control of debug interface")) {
			//	ImGuiIO& io = ImGui::GetIO();
			//	io.ConfigFlags -= ImGuiConfigFlags_NavEnableGamepad;
			//}
			}
	else {
		ImGui::Text("Joystick/Gamepad could not be found");
	}
	//ImGui::SliderInt("Left shoulder button action", &lbutton, 1, 16);
	//ImGui::Combo("Left shoulder button", &lbutton, controllercmds, IM_ARRAYSIZE(controllercmds));
ImGui::End();
}

//Log Window
if (logwindow)
{
ImGui::Begin("Log", &logwindow);
ImGui::Text("Log Level:");
ImGui::SameLine();
if (ImGui::Button("Errors Only"))
		log_debug_on(0);
ImGui::SameLine();
if (ImGui::Button("Standard"))
		log_debug_on(1);
ImGui::SameLine();
if (ImGui::Button("Verbose"))
		log_debug_on(2);
ImGui::SameLine();
ImGui::Checkbox("Coloured stdout", &dbg.debug_coltext);
ImGui::Separator();
ImGui::Text("Log:");
ImGui::SameLine();
ImGui::Checkbox("Meminfo", &dbg.debug_meminfo);
ImGui::SameLine();
ImGui::Checkbox("Boot", &dbg.debug_scripboot);
ImGui::SameLine();
ImGui::Checkbox("'Bad pic' warnings", &dbg.debug_badpic);
ImGui::SameLine();
ImGui::Checkbox("Returnint values", &dbg.debug_returnint);
ImGuiInputTextFlags input_text_flags = ImGuiInputTextFlags_EnterReturnsTrue | ImGuiInputTextFlags_CallbackHistory | ImGuiInputTextFlags_EscapeClearsAll;
//Todo: put this into a function
if (console_active)
{
		if (dinkc_enabled)
		{
				if (ImGui::InputText("DinkC Console", debugbuf, IM_ARRAYSIZE(debugbuf), input_text_flags, &DinkcConsoleCallback))
				{
				dinkc_execute_one_liner(debugbuf);
				strcpy(debugbuf, "");
				}
		}

		if (dinklua_enabled)
		{
				if (ImGui::InputText("Lua Console", luabuf, IM_ARRAYSIZE(luabuf), input_text_flags, &LuaConsoleCallback))
				{
				dinklua_execute_one_liner(luabuf);
				LuaHistoryPos = -1;
				LuaHistory.push_back(strdup(luabuf));
				strcpy(luabuf, "");
				}
		}
}
ImGui::Separator();
ImGui::Checkbox("Auto-scroll", &AutoScroll);
ImGui::SameLine();
if (ImGui::Button("Clear"))
		memset(&Buf, 0, sizeof(Buf));
ImGui::SameLine();
if (ImGui::Button("Clear console"))
	printf("\e[1;1H\e[2J");
ImGui::SameLine();
if (ImGui::Button("Copy last message to clipboard"))
{
		ImGui::LogToClipboard();
		ImGui::LogText("%s", log_getLastLog());
		ImGui::LogFinish();
}
ImGui::SameLine();
if (ImGui::Button("Copy all to clipboard")) {
	ImGui::LogToClipboard();
	ImGui::LogText(Buf.c_str());
	ImGui::LogFinish();
}
ImGui::Separator();
// Child window with text in it
ImGui::BeginChild("scrolling");
const char *buf = Buf.begin();
const char *buf_end = Buf.end();
ImGui::TextUnformatted(buf, buf_end);
if (AutoScroll)
		ImGui::SetScrollHereY(1.0f);

ImGui::EndChild();
ImGui::End();
}

//Ultimate Cheat
if (ultcheat) {
ImGui::Begin("Ultimater Cheat", &ultcheat);

if (ImGui::Button("Refill health")) {
    *plife = *plifemax;
	draw_status_all();
}
ImGui::SameLine();
if (ImGui::Button("Redraw status bar"))
    draw_status_all();
ImGui::SameLine();
if (ImGui::Button("Unlock screen"))
	screenlock = 0;
ImGui::SameLine();
if (ImGui::Button("Load/Draw screen")) {
	game_load_screen(g_dmod.map.loc[*pplayer_map]);
	draw_screen_game();
}
if (ImGui::Button("Kill enemies (no exp)")) {
	for (int i = 2; i < MAX_SPRITES_AT_ONCE; i++) {
		if (spr[i].hitpoints > 0)
		spr[i].damage = spr[i].hitpoints;
	}
}
ImGui::SameLine();
if (ImGui::Button("Unfreeze Sprite 1")) {
	spr[1].freeze = 0;
}

ImGui::Checkbox("Infinite health", &debug_infhp);
ImGui::SameLine();
ImGui::Checkbox("Infinite magic", &debug_magcharge);
ImGui::SameLine();
ImGui::Checkbox("Noclip", &debug_noclip);
ImGui::SameLine();
ImGui::Checkbox("Walk off screen", &debug_walkoff);
//ImGui::Checkbox("Screenlock inactive", &debug_screenlock);
//ImGui::SameLine();
//ImGui::Checkbox("Push", !!push_active);
ImGui::Separator();
ImGui::SliderInt("Gold", pgold, 0, 99000);
ImGui::SliderInt("Attack", pstrength, 0, 999);
ImGui::SliderInt("Defense", pdefense, 0, 999);
ImGui::SliderInt("Magic", pmagic, 0, 999);
ImGui::SliderInt("Life", plife, 0, *plifemax);
ImGui::SliderInt("Lifemax", plifemax, 1, 500);
ImGui::SliderInt("Level", plevel, 1, 32);
ImGui::SliderInt("Map Screen", pplayer_map, 1, 768);
ImGui::SliderInt("Walk Speed (lower is faster)", &dinkspeed, 0, 5);
ImGui::SliderInt("Screen X", &spr[1].x, 0, 620);
ImGui::SliderInt("Screen Y", &spr[1].y, -1, 400);

ImGui::End();
}

if (graphicsopt) {
//Rendering options - obsoleted
ImGui::Begin("Graphics", &graphicsopt);
ImGui::Text("Video driver: %s", SDL_GetCurrentVideoDriver());
ImGui::Separator();
ImGui::Checkbox("Draw all Collision Data", &drawhard);
//ImGui::Checkbox("Draw Sprites", &debug_drawsprites);
ImGui::Checkbox("Redraw background", &render_tiles);
if (ImGui::Button("Fade Down"))
dinkc_execute_one_liner("fade_down()");
if (ImGui::Button("Fade up instantly"))
up_cycle();
ImGui::End();
}

if (scripfo) {
ImGui::Begin("Scrips in use", &scripfo, ImGuiWindowFlags_AlwaysAutoResize);
//For making scrip to show just one time
int m = 0;
//C has no reflection apparently
if (ImGui::BeginTable("table2", 4, ImGuiTableFlags_Borders)) {
	ImGui::TableSetupColumn("No.");
	ImGui::TableSetupColumn("Name");
	ImGui::TableSetupColumn("Sprite");
	ImGui::TableSetupColumn("Data");
	ImGui::TableHeadersRow();
for (int i = 0; i < MAX_SCRIPTS; i++)
	if (sinfo[i] != NULL)
    {
		ImGui::TableNextColumn();
		ImGui::Text("%d", i);
		ImGui::TableNextColumn();
		if (ImGui::Button(sinfo[i]->name)) {
			myscript = i;
		}
		//ImGui::SameLine();
		//ImGui::PushID(i);
		//if (ImGui::Button("Kill")) {
		//	kill_script(i);
		//}
		//ImGui::PopID();
        //ImGui::Text("%s", rinfo[i]->name);
		ImGui::TableNextColumn();
		ImGui::Text("%d", sinfo[i]->sprite);
		ImGui::TableNextColumn();
		ImGui::Text("%s", sinfo[i]->data);
		m++;
    }
	ImGui::EndTable();
}

if (rinfo_code[myscript] != NULL)
ImGui::InputTextMultiline("##SelectedScript", rinfo_code[myscript], rinfo(myscript)->end, ImVec2(-1.0f,0.0f), ImGuiInputTextFlags_ReadOnly);
ImGui::Separator();
ImGui::Text("Scrips to show:");
ImGui::SameLine();

float progress = (float)m / (float)MAX_SCRIPTS;
char buf[32];
sprintf(buf, "%d/%d", m, MAX_SCRIPTS);
ImGui::ProgressBar(progress, ImVec2(0.0f, 0.0f), buf);

ImGui::End();
}
//Sprite Information
if (sprinfo) {
    ImGui::Begin("Sprite Lists", &sprinfo, ImGuiWindowFlags_AlwaysAutoResize);
	if (ImGui::BeginTabBar("spritab")) {
		if (ImGui::BeginTabItem("Live Sprites")) {
			//Sprite counter
			int m = 0;
			for (int i = 1; i < MAX_SPRITES_AT_ONCE; i++)
				if (spr[i].active == 1)
					m++;
			float progress = (float)m / (float)MAX_SPRITES_AT_ONCE;
			char buf[32];
			sprintf(buf, "%d/%d in use", m, MAX_SPRITES_AT_ONCE);
			ImGui::ProgressBar(progress, ImVec2(0.0f, 0.0f), buf);
			//ImGui::Checkbox("Filter editor sprites", &sprilivechk);
    	if (ImGui::BeginTable("table1", 19, ImGuiTableFlags_Borders | ImGuiTableFlags_Reorderable)) {
			ImGui::TableSetupColumn("Number");
			ImGui::TableSetupColumn("x");
			ImGui::TableSetupColumn("y");
			ImGui::TableSetupColumn("Seq");
			ImGui::TableSetupColumn("Frame");
			ImGui::TableSetupColumn("pSeq");
			ImGui::TableSetupColumn("pFrame");
			ImGui::TableSetupColumn("Timing");
			ImGui::TableSetupColumn("Script");
			ImGui::TableSetupColumn("Brain/parms");
			ImGui::TableSetupColumn("HP");
			ImGui::TableSetupColumn("Dir");
			ImGui::TableSetupColumn("Cue");
			ImGui::TableSetupColumn("Freeze");
			ImGui::TableSetupColumn("Size");
			ImGui::TableSetupColumn("Noclip");
			ImGui::TableSetupColumn("Editor No.");
			ImGui::TableSetupColumn("Hard");
			ImGui::TableSetupColumn("Nodraw");
			ImGui::TableHeadersRow();
			for (int x = 1; x < MAX_SPRITES_AT_ONCE; ++x) {
				if (spr[x].active && spr[x].brain != 8) {
					ImGui::TableNextColumn();
					ImGui::Text("%d", x);
					ImGui::SameLine();
					ImGui::PushID(x * 10000);
					if (ImGui::Button("Edit")) {
					spriedit = x;
					spreditors = true;
					flag = ImGuiTabItemFlags_SetSelected;
					}
					ImGui::PopID();
					ImGui::SameLine();
					ImGui::PushID(x * 1000000000);
					if (ImGui::Button("Kill"))
					spr[x].kill_ttl = 1;
					ImGui::PopID();
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].x);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].y);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].seq);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].frame);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].pseq);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].pframe);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].timing);
					ImGui::TableNextColumn();
					if (spr[x].script > 0)
					//ImGui::Text("%s", rinfo[spr[x].script]->name);
					ImGui::Text("%i", spr[x].script);
					else
					ImGui::Text("None");
					ImGui::TableNextColumn();
					ImGui::Text("%d/%d/%d", spr[x].brain, spr[x].brain_parm, spr[x].brain_parm2);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].hitpoints);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].dir);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].que);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].freeze);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].size);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].noclip);
					ImGui::TableNextColumn();
					//edit button
					ImGui::PushID(x * (100000));
					if (spr[x].sp_index > 0) {
					if (ImGui::Button("Edit")) {
					mspried = spr[x].sp_index;
					spreditors = true;
					mflag = ImGuiTabItemFlags_SetSelected;
					}
					ImGui::SameLine();
					}
					ImGui::PopID();
					ImGui::Text("%d", spr[x].sp_index);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].hard);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].nodraw);
				}
    	}
		ImGui::EndTable();
		}
		ImGui::EndTabItem();
		}

	//Text sprites
	if (ImGui::BeginTabItem("Text Sprites")) {
		if (ImGui::BeginTable("table2", 7, ImGuiTableFlags_Borders)) {
			ImGui::TableSetupColumn("Number");
			ImGui::TableSetupColumn("x");
			ImGui::TableSetupColumn("y");
			ImGui::TableSetupColumn("Strength");
			ImGui::TableSetupColumn("Defense");
			ImGui::TableSetupColumn("Parent sprite");
			ImGui::TableSetupColumn("Text");
			ImGui::TableHeadersRow();

			for (int x = 1; x < MAX_SPRITES_AT_ONCE; ++x) {
				if (spr[x].active && spr[x].brain == 8) {
					ImGui::TableNextColumn();
					ImGui::Text("%d", x);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].x);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].y);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].strength);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].defense);
					ImGui::TableNextColumn();
					ImGui::Text("%d", spr[x].text_owner);
					ImGui::TableNextColumn();
					ImGui::PushItemWidth(600);
					ImGui::Text(spr[x].text);
					ImGui::PopItemWidth();
				}
			}
			ImGui::EndTable();
		}
		ImGui::EndTabItem();
	}

	//Map Sprites
	if (mode < 2)
	ImGui::BeginDisabled();
	if (ImGui::BeginTabItem("Map Sprites")) {
		ImGui::Checkbox("Show vacant slots", &mspriunused);
		if (ImGui::BeginTable("table3", 23, ImGuiTableFlags_Borders)) {
			ImGui::TableSetupColumn("Number");
			ImGui::TableSetupColumn("x");
			ImGui::TableSetupColumn("y");
			ImGui::TableSetupColumn("Seq");
			ImGui::TableSetupColumn("Frame");
			ImGui::TableSetupColumn("Type");
			ImGui::TableSetupColumn("Size");
			ImGui::TableSetupColumn("In Use");
			ImGui::TableSetupColumn("Brain");
			ImGui::TableSetupColumn("Script");
			ImGui::TableSetupColumn("Speed");
			ImGui::TableSetupColumn("Timing");
			ImGui::TableSetupColumn("Cue");
			ImGui::TableSetupColumn("Hard");
			ImGui::TableSetupColumn("Warp on");
			ImGui::TableSetupColumn("Screen");
			ImGui::TableSetupColumn("x dest");
			ImGui::TableSetupColumn("y dest");
			ImGui::TableSetupColumn("Warp seq");
			ImGui::TableSetupColumn("Sound");
			ImGui::TableSetupColumn("Vision");
			ImGui::TableSetupColumn("Touch damage");
			ImGui::TableSetupColumn("Gold");
			ImGui::TableHeadersRow();

			for (int x = 1; x <= MAX_SPRITES_EDITOR; x++) {
				if (cur_ed_screen.sprite[x].active || mspriunused) {
					ImGui::TableNextColumn();
					ImGui::Text("%d", x);
					ImGui::SameLine();
					ImGui::PushID(x * 10);
					if (ImGui::Button("Edit")) {
						mspried = x;
						spreditors = true;
						mflag = ImGuiTabItemFlags_SetSelected;
					}
					ImGui::PopID();
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].x);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].y);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].seq);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].frame);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].type);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].size);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].active);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].brain);
					ImGui::TableNextColumn();
					ImGui::Text("%s", cur_ed_screen.sprite[x].script);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].speed);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].timing);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].que);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].hard);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].is_warp);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].warp_map);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].warp_x);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].warp_y);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].parm_seq);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].sound);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].vision);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].touch_damage);
					ImGui::TableNextColumn();
					ImGui::Text("%d", cur_ed_screen.sprite[x].gold);
				}
			}

			ImGui::EndTable();
		}
		ImGui::EndTabItem();
	}
	if (mode < 2)
		ImGui::EndDisabled();
	ImGui::EndTabBar();
	}
ImGui::End();
}

if (spreditors) {
	ImGui::Begin("Sprite Editor", &spreditors, ImGuiWindowFlags_AlwaysAutoResize);
	if (ImGui::BeginTabBar("edittab")) {

	//Sprite editor tab
	if (ImGui::BeginTabItem("Live Sprite Editor", NULL, flag)) {
		//Tab item flags for edit button clicking
		flag = ImGuiTabItemFlags_None;
		ImGui::Text("Activate the console to type directly into the boxes");
		ImGui::Text("Selected Sprite: %d", spriedit);
		ImGui::SameLine();
		ImGui::PushButtonRepeat(true);
		if (ImGui::ArrowButton("##left", ImGuiDir_Left) && spriedit != 1) { spriedit--; }
		ImGui::SameLine(0.0f, ImGui::GetStyle().ItemInnerSpacing.x);
		if (ImGui::ArrowButton("##right", ImGuiDir_Right)) { spriedit++; }
		ImGui::PopButtonRepeat();
		ImGui::SameLine();
		ImGui::Checkbox("Highlight", &spoutline);
		ImGui::Separator();
		//Sprite properties
		if (spr[spriedit].active) {
		if (ImGui::Button("Deactivate sprite")) {
			spr[spriedit].active = 0;
			}
		}
		else {
			if (ImGui::Button("Activate sprite")) {
				spr[spriedit].active = 1;
			}
		}
		if (spr[spriedit].sp_index > 0) {
			ImGui::SameLine();
			if (ImGui::Button("Edit corresponding map sprite")) {
			mspried = spr[spriedit].sp_index;
			mflag = ImGuiTabItemFlags_SetSelected;
			}
			if (ImGui::Button("Copy and save live sprite data to map sprite")) {
				mspried = spr[spriedit].sp_index;
				livesprtomap(spriedit);
				save_screen(g_dmod.map.map_dat.c_str(), g_dmod.map.loc[*pplayer_map]);
			}
		}
		ImGui::SliderInt("X", &spr[spriedit].x, 0, xmax);
		ImGui::SliderInt("Y", &spr[spriedit].y, 0, ymax);
		//ImGui::SliderInt("mx", &spr[spriedit].mx, 0, xmax);
		//ImGui::SliderInt("my", &spr[spriedit].my, 0, ymax);
		//ImGui::Text("Moveman: %d", spr[spriedit].moveman);
		if (ImPlot::BeginPlot("Screen Positioning", ImVec2(xmax / 2, ymax / 2), ImPlotFlags_Crosshairs | ImPlotFlags_NoLegend)) {
			ImPlot::SetupAxesLimits(0, xmax, 0, ymax);
			ImPlot::SetupAxis(ImAxis_Y1, NULL, ImPlotAxisFlags_Invert | ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin | ImPlotAxisFlags_NoTickLabels);
			ImPlot::SetupAxis(ImAxis_X1, NULL, ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin | ImPlotAxisFlags_NoTickLabels);
			ImPlot::PlotScatter("Selected sprite", &spr[spriedit].x, &spr[spriedit].y, 1, ImPlotScatterFlags_NoClip);

			if (ImPlot::IsPlotHovered() && ImGui::IsMouseClicked(0)) {
            ImPlotPoint pt = ImPlot::GetPlotMousePos();
			if (ImGui::GetIO().KeyCtrl) {
            spr[spriedit].x = pt.x;
			} else if (ImGui::GetIO().KeyShift)
			{
			spr[spriedit].y = pt.y;
			}
			else {
				spr[spriedit].x = pt.x;
				spr[spriedit].y = pt.y;
			}
        }
			ImPlot::EndPlot();
		}
		if (ImGui::Button("Copy X and Y values to clipboard")) {
			ImGui::LogToClipboard();
			ImGui::LogText("%d, %d", spr[spriedit].x, spr[spriedit].y);
			ImGui::LogFinish();
		}
		ImGui::Separator();
		ImGui::RadioButton("Frozen", &spr[spriedit].freeze, 1);
		ImGui::SameLine();
		ImGui::RadioButton("Thawed", &spr[spriedit].freeze, 0);
		ImGui::RadioButton("Not nohit", &spr[spriedit].nohit, 0);
		ImGui::SameLine();
		ImGui::RadioButton("Nohit active", &spr[spriedit].nohit, 1);
		ImGui::RadioButton("Drawn", &spr[spriedit].nodraw, 0);
		ImGui::SameLine();
		ImGui::RadioButton("Nodraw", &spr[spriedit].nodraw, 1);
		ImGui::RadioButton("Hard on", &spr[spriedit].hard, 0);
		ImGui::SameLine();
		ImGui::RadioButton("Hard off", &spr[spriedit].hard, 1);
		ImGui::SliderInt("pSequence", &spr[spriedit].pseq, 1, 1000);
		ImGui::SliderInt("pFrame", &spr[spriedit].pframe, 1, 69);
		ImGui::SliderInt("Direction", &spr[spriedit].dir, 1, 8);
		ImGui::SliderInt("Size", &spr[spriedit].size, 1, 200);
		ImGui::SameLine();
		if (ImGui::SmallButton("Reset"))
			spr[spriedit].size = 100;
		ImGui::InputInt("Timing", &spr[spriedit].timing);
		ImGui::InputInt("Frame Delay", &spr[spriedit].frame_delay);
		//Change to drop down box eventually
		//ImGui::InputInt("Brain", &spr[spriedit].brain, 1, 2);
		ImGui::SliderInt("Depth Cue", &spr[spriedit].que, -1000, 1000);
		ImGui::InputInt("Sound", &spr[spriedit].sound);
		ImGui::InputText("Text", spr[spriedit].text, sizeof(spr[spriedit].text));
		//ImGui::SliderInt("Hitpoints", &spr[spriedit].hitpoints, 0, 100);
		ImGui::Separator();
		if (ImGui::TreeNode("More sequences")) {
			ImGui::InputInt("Sequence", &spr[spriedit].seq);
			ImGui::InputInt("Frame", &spr[spriedit].frame);
			ImGui::InputInt("Blood Sequence", &spr[spriedit].bloodseq);
			ImGui::InputInt("Blood number", &spr[spriedit].bloodnum);
			ImGui::InputInt("Base Walk", &spr[spriedit].base_walk);
			ImGui::InputInt("Base Idle", &spr[spriedit].base_idle);
			ImGui::InputInt("Base Hit", &spr[spriedit].base_hit);
			ImGui::InputInt("Base Attack", &spr[spriedit].base_attack);
			ImGui::InputInt("Base Die", &spr[spriedit].base_die);
			ImGui::TreePop();
		}
		ImGui::Separator();
		if (ImGui::TreeNode("Enemy parameters")) {
			ImGui::InputInt("Hitpoints", &spr[spriedit].hitpoints);
			ImGui::InputInt("Touch damage", &spr[spriedit].touch_damage);
			ImGui::InputInt("Defence", &spr[spriedit].defense);
			ImGui::InputInt("Strength", &spr[spriedit].strength);
			ImGui::InputInt("Experience points", &spr[spriedit].exp);
			ImGui::InputInt("Attack hit sound", &spr[spriedit].attack_hit_sound);
			ImGui::InputInt("Hit sound Hz", &spr[spriedit].attack_hit_sound_speed);
			ImGui::TreePop();
		}
		ImGui::Separator();
		if (ImGui::TreeNode("Brain and parms")) {
			ImGui::InputInt("Brain", &spr[spriedit].brain);
			ImGui::InputInt("Parm 1", &spr[spriedit].brain_parm);
			ImGui::InputInt("Parm 2", &spr[spriedit].brain_parm2);
			ImGui::TreePop();
		}
		ImGui::Separator();
		if (ImGui::TreeNode("Trimming")) {
			if (ImGui::Button("Set to bounding box")) {
				spr[spriedit].alt.right = k[getpic(spriedit)].box.right;
				spr[spriedit].alt.bottom = k[getpic(spriedit)].box.bottom;
			}
			ImGui::SliderInt("Trim top", &spr[spriedit].alt.top, 0, k[getpic(spriedit)].box.bottom);
			ImGui::SliderInt("Trim left", &spr[spriedit].alt.left, 0, k[getpic(spriedit)].box.right);
			//ImGui::InputInt("Trim left", &spr[spriedit].alt.left);
			//ImGui::InputInt("Trim/show right", &spr[spriedit].alt.right);
			ImGui::SliderInt("Trim/show right", &spr[spriedit].alt.right, 0, k[getpic(spriedit)].box.right);
			//ImGui::InputInt("Trim/show bottom", &spr[spriedit].alt.bottom);
			ImGui::SliderInt("Trim/show bottom", &spr[spriedit].alt.bottom, 0, k[getpic(spriedit)].box.bottom);
			//TODO: add easy trimming feature with implot
			//Get the bounding box of the sprite
			if (ImPlot::BeginPlot("Drag to trim", ImVec2(320, 240), ImPlotFlags_NoLegend)) {
				ImPlot::SetupAxesLimits(0, k[getpic(spriedit)].box.right, 0, k[getpic(spriedit)].box.bottom);
				//invert axis because dink is dumb
				ImPlot::SetupAxis(ImAxis_Y1, NULL, ImPlotAxisFlags_Invert);
				trimbox.X.Min = spr[spriedit].alt.left;
				trimbox.X.Max = spr[spriedit].alt.right;
				trimbox.Y.Min = spr[spriedit].alt.top;
				trimbox.Y.Max = spr[spriedit].alt.bottom;
				ImPlot::DragRect(0, &trimbox.X.Min, &trimbox.Y.Min, &trimbox.X.Max, &trimbox.Y.Max,ImVec4(1,0,1,1), ImPlotDragToolFlags_Delayed);
				//Set trimming with dragging
				spr[spriedit].alt.left = trimbox.X.Min;
				spr[spriedit].alt.right = trimbox.X.Max;
				spr[spriedit].alt.top = trimbox.Y.Min;
				spr[spriedit].alt.bottom = trimbox.Y.Max;
				ImPlot::EndPlot();
			}
			ImGui::TreePop();
		}
		ImGui::Separator();
		if (ImGui::TreeNode("Dink.ini properties")) {
			ImGui::Text("These values are not persistent after editing");
			ImGui::Checkbox("Hardbox preview", &hboxprev);
			ImGui::SameLine();
			ImGui::Checkbox("Dot preview", &ddotprev);
			if (ImGui::Button("Set all to zero")) {
				memset(&k[getpic(spriedit)].hardbox, 0, sizeof(pic_info));
				k[getpic(spriedit)].xoffset = 0;
				k[getpic(spriedit)].yoffset = 0;
			}
			ImGui::SeparatorText("Hardbox");
			ImGui::TextWrapped("Be aware that these may have been edited by an init() line at some point.");
			ImGui::InputInt("Left", &k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.left);
			ImGui::InputInt("Top", &k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.top);
			ImGui::InputInt("Right", &k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.right);
			ImGui::InputInt("Bottom", &k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.bottom);
			ImGui::SeparatorText("Depth dot offset");
			ImGui::InputInt("X", &k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].xoffset);
			ImGui::InputInt("Y", &k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].yoffset);
			ImGui::Separator();
			//ImGui::Text("Box is %d right %d bottom", k[getpic(spriedit)].box.right, k[getpic(spriedit)].box.bottom);

			if (spr[spriedit].brain != 8) {
				ImGui::Checkbox("Interactive hardbox editing", &sprplotedit);
				ImGui::Text("Ctrl+Click to adjust dot");
			if (ImPlot::BeginPlot(("Box and dot"), ImVec2(320, 240), ImPlotFlags_NoLegend)) {
				ImPlot::SetupAxesLimits(0, k[getpic(spriedit)].box.right, 0, k[getpic(spriedit)].box.bottom);
				ImPlot::SetupAxis(ImAxis_Y1, NULL, ImPlotAxisFlags_Invert);
				//ImPlot::SetupAxis(ImAxis_X1, NULL, ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin);
				ImPlot::PlotScatter("Dot", &k[getpic(spriedit)].xoffset, &k[getpic(spriedit)].yoffset, 1, ImPlotScatterFlags_NoClip);
				hbox.X.Min = k[getpic(spriedit)].xoffset + k[getpic(spriedit)].hardbox.left;
				hbox.X.Max = k[getpic(spriedit)].xoffset + k[getpic(spriedit)].hardbox.right;
				hbox.Y.Min = k[getpic(spriedit)].yoffset + k[getpic(spriedit)].hardbox.top;
				hbox.Y.Max = k[getpic(spriedit)].yoffset + k[getpic(spriedit)].hardbox.bottom;
				ImPlot::DragRect(0, &hbox.X.Min, &hbox.Y.Min, &hbox.X.Max, &hbox.Y.Max,ImVec4(1,0,1,1), ImPlotDragToolFlags_Delayed);
				if (sprplotedit) {
				k[getpic(spriedit)].hardbox.left = hbox.X.Min - k[getpic(spriedit)].xoffset;
				k[getpic(spriedit)].hardbox.right = hbox.X.Max - k[getpic(spriedit)].xoffset;
				k[getpic(spriedit)].hardbox.top = hbox.Y.Min - k[getpic(spriedit)].yoffset;
				k[getpic(spriedit)].hardbox.bottom = hbox.Y.Max - k[getpic(spriedit)].yoffset;
				}
				if (ImPlot::IsPlotHovered() && ImGui::IsMouseClicked(0) && ImGui::GetIO().KeyCtrl) {
            		ImPlotPoint pt = ImPlot::GetPlotMousePos();
            		k[getpic(spriedit)].xoffset = pt.x;
					k[getpic(spriedit)].yoffset = pt.y;
        		}
				ImPlot::EndPlot();
			}
			}
			ImGui::Separator();
			if (ImGui::Button("Copy depth dot and hardbox values to clipboard")) {
				ImGui::LogToClipboard();
				ImGui::LogText("%d %d %d %d %d %d", k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].xoffset, k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].yoffset,
				k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.left, k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.top, k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.right, k[seq[spr[spriedit].pseq].frame[spr[spriedit].pframe]].hardbox.bottom);
				ImGui::LogFinish();
			}
			if (ImGui::Button("Append SET_SPRITE_INFO line to Dink.ini")) {
				add_text();
		}
		ImGui::TreePop();
		}
		ImGui::Separator();
		if (ImGui::Button("Copy create_sprite() command to clipboard")) {
			ImGui::LogToClipboard();
			ImGui::LogText("create_sprite(%d, %d, %d, %d, %d)", spr[spriedit].x, spr[spriedit].y, spr[spriedit].brain, spr[spriedit].pseq, spr[spriedit].pframe);
			ImGui::LogFinish();
		}
		ImGui::EndTabItem();
	} // Start of screen sprite editor
	if (ImGui::BeginTabItem("Map Sprite Editor", nullptr, mflag)) {
		mflag = ImGuiTabItemFlags_None;
		if (mode == 1)
		ImGui::BeginDisabled();
		ImGui::InputText("Screen Script", cur_ed_screen.script, sizeof(cur_ed_screen.script));
		if (ImGui::Button("Remove all sprites on this screen")) {
			memset(&cur_ed_screen.sprite, 0, sizeof(cur_ed_screen.sprite));
		}
		ImGui::Separator();
		ImGui::Spacing();
		int maxframe = 0;
		//ImGui::Text("Careful now.");
		//Index selector
		ImGui::Text("Selected editor sprite: %d", mspried);
		ImGui::SameLine();
		ImGui::PushButtonRepeat(true);
		if (ImGui::ArrowButton("##back", ImGuiDir_Left)) {if (mspried > 1) mspried--;}
		ImGui::SameLine(0.0f, ImGui::GetStyle().ItemInnerSpacing.x);
		if (ImGui::ArrowButton("##forward", ImGuiDir_Right)) { mspried++; }
		ImGui::PopButtonRepeat();
		ImGui::SameLine();
		if (ImGui::Button("1"))
			mspried = 1;
		if (ImGui::IsItemHovered())
			ImGui::SetTooltip("Return to the first sprite index");
		//Various buttons
		if (ImGui::Button("Redraw screen")) {
			draw_screen_game();
		}
		ImGui::Separator();

		if (cur_ed_screen.sprite[mspried].active == 0) {
		if (ImGui::Button("Activate Sprite")) {
			cur_ed_screen.sprite[mspried].active = 1;
		}
		}
		else {
		if (ImGui::Button("Remove Sprite"))
			cur_ed_screen.sprite[mspried].active = 0;

		ImGui::SameLine();
		if (ImGui::Button("Make duplicate")) {
			//Find an empty slot
			for (int i = 1; i < MAX_SPRITES_EDITOR; i++) {
				if (cur_ed_screen.sprite[i].active == 0) {
					memcpy(&cur_ed_screen.sprite[i], &cur_ed_screen.sprite[mspried], sizeof(editor_sprite));
					mspried = i;
					break;
				}
			}
		}
		}
		ImGui::Separator();
		//ImGui::TextWrapped("Some of these values may have been overwritten by editor_frame() etc");
		if (play.spmap[*pplayer_map].type[mspried] != 0 || play.spmap[*pplayer_map].frame[mspried] != 0)
		ImGui::TextColored(ImVec4(1.0f,0.0f,0.0f,1.0f), "Warning: values overwritten by editor_frame() etc");
		//Start of parameters
		ImGui::InputText("Script", cur_ed_screen.sprite[mspried].script, sizeof(cur_ed_screen.sprite[mspried].script));
		ImGui::Separator();
		ImGui::Spacing();
		ImGui::SliderInt("X position", &cur_ed_screen.sprite[mspried].x, 0, xmax);
		ImGui::SliderInt("Y position", &cur_ed_screen.sprite[mspried].y, 0, ymax);
		if (ImPlot::BeginPlot("Positioning", ImVec2(xmax / 2, ymax / 2), ImPlotFlags_Crosshairs| ImPlotFlags_NoLegend)) {
			ImPlot::SetupAxesLimits(0, xmax, 0, ymax);
			ImPlot::SetupAxis(ImAxis_Y1, NULL, ImPlotAxisFlags_Invert | ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin);
			ImPlot::SetupAxis(ImAxis_X1, NULL, ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin);
			ImPlot::PlotScatter("Selected sprite", &cur_ed_screen.sprite[mspried].x, &cur_ed_screen.sprite[mspried].y, 1, ImPlotScatterFlags_NoClip);

			if (ImPlot::IsPlotHovered() && ImGui::IsMouseClicked(0)) {
            ImPlotPoint pt = ImPlot::GetPlotMousePos();
			if (ImGui::GetIO().KeyCtrl) {
            cur_ed_screen.sprite[mspried].x = pt.x;
			} else if (ImGui::GetIO().KeyShift)
			{
			cur_ed_screen.sprite[mspried].y = pt.y;
			}
			else {
				cur_ed_screen.sprite[mspried].x = pt.x;
				cur_ed_screen.sprite[mspried].y = pt.y;
			}
        }
			ImPlot::EndPlot();
		}
		ImGui::RadioButton("Type 0: background", &cur_ed_screen.sprite[mspried].type, 0);
		ImGui::SameLine();
		ImGui::RadioButton("1: normal", &cur_ed_screen.sprite[mspried].type, 1);
		ImGui::SameLine();
		ImGui::RadioButton("2: invisible", &cur_ed_screen.sprite[mspried].type, 2);
		ImGui::RadioButton("Hard off", &cur_ed_screen.sprite[mspried].hard, 1);
		ImGui::SameLine();
		ImGui::RadioButton("Hard on", &cur_ed_screen.sprite[mspried].hard, 0);
		ImGui::RadioButton("Nohit off", &cur_ed_screen.sprite[mspried].nohit, 0);
		ImGui::SameLine();
		ImGui::RadioButton("Nohit on", &cur_ed_screen.sprite[mspried].nohit, 1);
		ImGui::SliderInt("Vision", &cur_ed_screen.sprite[mspried].vision, 0, 9);
		ImGui::InputInt("Sequence", &cur_ed_screen.sprite[mspried].seq,1);
		ImGui::SliderInt("Frame", &cur_ed_screen.sprite[mspried].frame, 1, 50);
		ImGui::SliderInt("Size", &cur_ed_screen.sprite[mspried].size, 1, 200);
		ImGui::SameLine();
		if (ImGui::SmallButton("Reset"))
			cur_ed_screen.sprite[mspried].size = 100;
		ImGui::InputInt("Brain", &cur_ed_screen.sprite[mspried].brain);
		ImGui::SliderInt("Depth Cue", &cur_ed_screen.sprite[mspried].que, -100, 100);
		ImGui::Separator();
		ImGui::InputInt("Speed", &cur_ed_screen.sprite[mspried].speed);
		ImGui::InputInt("Timing", &cur_ed_screen.sprite[mspried].timing);
		ImGui::Separator();
		if (ImGui::TreeNode("Warp properties")) {
			ImGui::RadioButton("Warp Off", &cur_ed_screen.sprite[mspried].is_warp, 0);
			ImGui::SameLine();
			ImGui::RadioButton("Warp On", &cur_ed_screen.sprite[mspried].is_warp, 1);
			ImGui::InputInt("Map screen", &cur_ed_screen.sprite[mspried].warp_map);
			ImGui::SliderInt("X position", &cur_ed_screen.sprite[mspried].warp_x, 0, xmax);
			ImGui::SliderInt("Y position", &cur_ed_screen.sprite[mspried].warp_y, 0, ymax);
			ImGui::InputInt("Warp sequence", &cur_ed_screen.sprite[mspried].parm_seq);
			ImGui::InputInt("Sound", &cur_ed_screen.sprite[mspried].sound);
			ImGui::TreePop();
		}
		if (ImGui::TreeNode("Enemy properties")) {
			ImGui::InputInt("Hitpoints", &cur_ed_screen.sprite[mspried].hitpoints);
			ImGui::InputInt("Strength", &cur_ed_screen.sprite[mspried].strength);
			ImGui::InputInt("Defence", &cur_ed_screen.sprite[mspried].defense);
			ImGui::InputInt("Experience points", &cur_ed_screen.sprite[mspried].exp);
			ImGui::InputInt("Touch damage", &cur_ed_screen.sprite[mspried].touch_damage);
			ImGui::InputInt("Gold", &cur_ed_screen.sprite[mspried].gold);
			ImGui::TreePop();
		}
		if (ImGui::TreeNode("Base sequences")) {
			ImGui::InputInt("Base walk", &cur_ed_screen.sprite[mspried].base_walk);
			ImGui::InputInt("Base attack", &cur_ed_screen.sprite[mspried].base_attack);
			ImGui::InputInt("Base idle", &cur_ed_screen.sprite[mspried].base_idle);
			ImGui::InputInt("Base die", &cur_ed_screen.sprite[mspried].base_die);
			ImGui::InputInt("Base hit (unused)", &cur_ed_screen.sprite[mspried].base_hit);
			ImGui::TreePop();
		}
		if (ImGui::TreeNode("Trimming")) {
			ImGui::InputInt("Trim top", &cur_ed_screen.sprite[mspried].alt.top);
			ImGui::InputInt("Trim left", &cur_ed_screen.sprite[mspried].alt.left);
			ImGui::InputInt("Trim/show right", &cur_ed_screen.sprite[mspried].alt.right);
			ImGui::InputInt("Trim/show bottom", &cur_ed_screen.sprite[mspried].alt.bottom);
			ImGui::TreePop();
		}
		ImGui::Separator();
		if (ImGui::Button("Save changes to MAP.DAT")) {
			save_screen(g_dmod.map.map_dat.c_str(), g_dmod.map.loc[*pplayer_map]);
		}


		if (mode == 1)
		ImGui::EndDisabled();

		ImGui::EndTabItem();
	}
	if (ImGui::BeginTabItem("Plot")) {
		if (ImPlot::BeginPlot("Sprites", ImVec2(640, 480), ImPlotFlags_NoLegend)) {
			ImPlot::SetupAxesLimits(0, xmax, 0, ymax);
			ImPlot::SetupAxis(ImAxis_Y1, NULL, ImPlotAxisFlags_Invert | ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin | ImPlotAxisFlags_NoTickLabels);
			ImPlot::SetupAxis(ImAxis_X1, NULL, ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin);
			ImPlot::PlotScatter("DinkPlot", &spr[1].x, spr[1].y, 1, ImPlotScatterFlags_NoClip);
			for (int i = 2; i < MAX_SPRITES_AT_ONCE; i++) {
				if (spr[i].active && spr[i].brain != 8) {
					ImPlot::PlotScatter("##sprite", &spr[i].x, &spr[i].y, 1, ImPlotScatterFlags_NoClip);
				}
			}
			ImPlot::EndPlot();
		}
		ImGui::EndTabItem();
	}
	ImGui::EndTabBar();
	}
	ImGui::End();
}


if (variableinfo) {
    ImGui::Begin("Variables", &variableinfo, ImGuiWindowFlags_AlwaysAutoResize);
	//editor and option(s)
	//ImGui::SeparatorText("Edit");
	//ImGui::Text("Variable: %s", play.var[varedit].name);
	//ImGui::SameLine();
	//ImGui::SetNextItemWidth(20.0f);
	//ImGui::DragInt("##var", &varedit, 1.0f, 1, MAX_VARS);
	//ImGui::InputInt("Value", &play.var[varedit].var);
	//ImGui::SameLine();
	//if (ImGui::SmallButton("Reset"))
	//play.var[varedit].var = 0;
	//ImGui::Spacing();
	int m = 0;
	int i;
	for (i = 1; i < MAX_VARS; i++)
		if (play.var[i].active == 1)
			m++;
	float progress = (float)m / (float)MAX_VARS;
	char buf[32];
	sprintf(buf, "%d/%d vars in use", m, MAX_VARS);
	ImGui::ProgressBar(progress, ImVec2(0.0f, 0.0f), buf);
	ImGui::Checkbox("Hide global variables", &globinfo);
	//Table begins
	if (ImGui::BeginTable("table5", 5, ImGuiTableFlags_Borders)) {
	ImGui::TableSetupColumn("No.");
	ImGui::TableSetupColumn("Name");
	ImGui::TableSetupColumn("Value");
	ImGui::TableSetupColumn("Active");
	ImGui::TableSetupColumn("Scope");
	ImGui::TableHeadersRow();
    for (int i = 1; i < MAX_VARS; i++) {
		if (!strncmp(play.var[i].name, "&", 1)) {
			if (play.var[i].scope == 0 && !globinfo)
			{
		ImGui::TableNextColumn();
		ImGui::Text("%d", i);
		ImGui::TableNextColumn();
		ImGui::Text("%s", play.var[i].name);
        //ImGui::Text(play.var[i].name);
		ImGui::TableNextColumn();
		//ImGui::Text("%d", play.var[i].var);
		ImGui::PushID(i * 100);
		ImGui::SetNextItemWidth(100);
		ImGui::InputInt("##value", &play.var[i].var);
		ImGui::PopID();
		ImGui::TableNextColumn();
		ImGui::Text("%d", play.var[i].active);
		ImGui::TableNextColumn();
		ImGui::Text("%d", play.var[i].scope);
		//if (i == varedit)
		//ImGui::TableSetBgColor(ImGuiTableBgTarget_RowBg0, ImGui::GetColorU32(ImVec4(0.7f, 0.3f, 0.3f, 0.65f)));
		ImGui::TableNextRow();
			}
		}

    }
	ImGui::EndTable();
	}
    ImGui::End();
}


if (jukebox) {
	ImGui::Begin("Jukebox", &jukebox, ImGuiWindowFlags_AlwaysAutoResize);
	//info
	if (something_playing()) {
		ImGui::Text("Title: %s", tag_title);
		ImGui::Text("Artist: %s", tag_artist);
		ImGui::Text("Album: %s", tag_album);
		ImGui::Text("Duration: %1.2f seconds", mus_duration);

		muspos = Mix_GetMusicPosition(NULL);
		ImGui::BeginDisabled();
		ImGui::SliderScalar("##Position", ImGuiDataType_Double, &muspos, &zero, &mus_duration);
		ImGui::EndDisabled();
	}
		else {
			ImGui::Text("Nothing is playing at the moment");
		}
		//if (mus_type == tracker)
		//if (ImGui::SliderInt("Set tracker pattern", &modorder, 0, 128))
		//Mix_ModMusicJumpToOrder(modorder);
	//Main box
	if (ImGui::BeginListBox("##juke"))
	{
	for (int n = 0; n < jlist.Size; n++)
	{
		const bool is_selected = (juker == n);
		if (jlist[n] != NULL && ImGui::Selectable(jlist[n], is_selected))
		{
			juker = n;
			PlayMidi(jlist[n]);
		}

		if (is_selected)
            ImGui::SetItemDefaultFocus();
	}
	ImGui::EndListBox();
	}

	//if (ImGui::Button("Refresh list"))
	//	jukelist();

	ImGui::End();
}

if (sfxinfo) {
    ImGui::Begin("Audio Settings", &sfxinfo, ImGuiWindowFlags_AlwaysAutoResize);
	dbg.sfxvol = gSoloud.getGlobalVolume();
	dbg.musvol = Mix_VolumeMusic(-1);
	//ImGui::Separator();
	//ImGui::SliderInt("Mod/XM order", &modorder, 0, 99);
	//ImGui::SameLine();
	//if (ImGui::Button("Set")) {
	//Mix_ModMusicJumpToOrder(modorder);
	//}
	ImGui::SeparatorText("Volume");
	if (ImGui::SliderFloat("SFX", &dbg.sfxvol, 0, 2.0f))
	gSoloud.setGlobalVolume(dbg.sfxvol);
	if (ImGui::SliderInt("Music", &dbg.musvol, 0, MIX_MAX_VOLUME))
	Mix_VolumeMusic(dbg.musvol);
	ImGui::SliderFloat("Speech", &dbg.speevol, 0.5f, 6.0f);
	ImGui::Checkbox("Speech synthesis narration of text", &dbg.debug_speechsynth);
	ImGui::Spacing();

	ImGui::SeparatorText("Background Music");
		//if (ImGui::InputText("Media file", midichlorian, IM_ARRAYSIZE(midichlorian), ImGuiInputTextFlags_EnterReturnsTrue)) {
		//PlayMidi(midichlorian);
		//memset(midichlorian, 0, 30);
		//modorder = 0;
		//}
		//ImGui::SameLine();
		//if (ImGui::SmallButton("Play")) {
		//	PlayMidi(midichlorian);
		//	memset(midichlorian, 0, 30);
		//}
		if (something_playing()) {
		ImGui::Text("Title: %s", tag_title);
		ImGui::Text("Artist: %s", tag_artist);
		ImGui::Text("Album: %s", tag_album);
		//ImGui::Text("Position: %1.2f / %1.2f", Mix_GetMusicPosition(NULL), Mix_MusicDuration(NULL));
		ImGui::Text("Duration: %1.2f seconds", mus_duration);
	}
		else {
			ImGui::Text("Nothing is playing at the moment");
		}
		if (mus_type == tracker)
		if (ImGui::SliderInt("Set tracker pattern", &modorder, 0, 128))
		Mix_ModMusicJumpToOrder(modorder);
	if (something_playing()) {
		if (ImGui::Button("Halt BGM"))
		bgm_quit();

		ImGui::SameLine();
		if (ImGui::Button("Rewind BGM"))
		Mix_RewindMusic();
		ImGui::SameLine();
	if (ImGui::Button("Pause BGM"))
		PauseMidi();
	}
	if (Mix_PausedMusic()) {
		ImGui::SameLine();
	if (ImGui::Button("Resume BGM"))
		ResumeMidi();
	}
	ImGui::Spacing();
	ImGui::SeparatorText("Sound Effects");
	float *buf = gSoloud.getWave();
	float *fft = gSoloud.calcFFT();
	if (dbg.devmode) {
	if (ImGui::TreeNode("Speech parameters")) {
		ImGui::SliderInt("Frequency", &spfreq, 100, 8000);
		ImGui::SliderInt("Declination", &spdec, 1, 10);
		ImGui::SliderInt("Speed", &spspeed, 1, 20);
		ImGui::SliderInt("Waveform", &spwave, 0, 6);
		if (ImGui::Button("Set parameters")) {
			speech.setParams(spfreq, spspeed, spdec, spwave);
		}
		ImGui::TreePop();
	}
	}
	ImGui::Checkbox("Direction-based positional audio", &dbg.debug_audiopos);
	ImGui::Checkbox("Positional audio distance delay", &dbg.audiodelay);
	//ImGui::SameLine(); HelpMarker("Attempts to simulate sounds at a distance. Must restart to take effect in most cases.");
	if (dbg.devmode) {
	if (ImGui::TreeNode("Distance attenuation parameters")) {
		ImGui::SliderFloat("Attenuation", &atten, 0.0001f, 2.0f);
		ImGui::SliderFloat("Distance Minimum", &dist_min, 0.01f, 50.0f);
		ImGui::SliderFloat("Distance Max", &dist_max, 1.0f, 700.0f);
		ImGui::SliderInt("Model", &dist_model, 0, 3);
		if (dist_model == 0) {
			ImGui::Text("No attenuation");
		} else if (dist_model == 1) {
			ImGui::Text("Inverse distance model");
		} else if (dist_model == 2) {
			ImGui::Text("Linear");
		}
		ImGui::TreePop();
	}
	}
	ImGui::PlotLines("##Wave", buf, 256, 0, "Wave", -1, 1, ImVec2(-1.0, 80));
	ImGui::PlotHistogram("##FFT", fft, 256 / 2, 0, "FFT", 0, 10, ImVec2(-1.0, 80), 8);
	//Progress bar
	int m = gSoloud.getActiveVoiceCount();
	float voices = (float)m / (float)NUM_CHANNELS;
	char buff[32];
	sprintf(buff, "%d/%d Active Voices", m, NUM_CHANNELS);
	ImGui::ProgressBar(voices, ImVec2(-1.0f, 0.0f), buff);
	if (ImGui::Button("Halt all SFX")) {
		halt_all_sounds();
	}
	ImGui::Spacing();

	//Only show in developer mode
	if (dbg.devmode) {
	ImGui::SeparatorText("Audio decoders currently in use");
	if (Mix_GetNumMusicDecoders() == 0) {
		ImGui::Text("No music decoders enabled. Sound is probably switched off.");
	}
	for (int i=0; i < Mix_GetNumMusicDecoders(); i++) {
		ImGui::Text("%s ", Mix_GetMusicDecoder(i));
		if (i % 4)
		ImGui::SameLine();
	}
	ImGui::Spacing();
	}
	//MixerX-specific stuff. Should probably make a define thing for standard mixer
	#ifdef HAVE_MIXER_EXT
	ImGui::SeparatorText("MixerX Settings");
	if (Mix_GetSoundFonts() != NULL) {
		ImGui::Text("SoundFont in use: %s", Mix_GetSoundFonts());
	}
	else {
	ImGui::InputText("SoundFont Path", sfbuf, IM_ARRAYSIZE(sfbuf));
	ImGui::SameLine();
	if (ImGui::SmallButton("Set"))
	{
	Mix_SetSoundFonts(sfbuf);
	}
	}



	if (ImGui::Button("ADLMIDI"))
		Mix_SetMidiPlayer(0);
	ImGui::SameLine();
	if (ImGui::Button("OPNMIDI"))
		Mix_SetMidiPlayer(3);
	ImGui::SameLine();
	if (ImGui::Button("FluidSynth"))
		Mix_SetMidiPlayer(4);
	ImGui::SameLine();
	if (ImGui::Button("TiMidity")) {
		Mix_SetMidiPlayer(2);
	}
	ImGui::SameLine();
	if (ImGui::Button("System")) {
		Mix_SetMidiPlayer(1);
	}
	ImGui::Separator();
	ImGui::Text("ADLMIDI Emulator");
	if (ImGui::Button("Nuked 1.8")) {
		Mix_ADLMIDI_setEmulator(ADLMIDI_OPL3_EMU_NUKED);
	}
	ImGui::SameLine();
	if (ImGui::Button("Nuked 1.7.4")) {
		Mix_ADLMIDI_setEmulator(ADLMIDI_OPL3_EMU_NUKED_1_7_4);
	}
	ImGui::SameLine();
	if (ImGui::Button("DOSBox")) {
		Mix_ADLMIDI_setEmulator(ADLMIDI_OPL3_EMU_DOSBOX);
	}
	ImGui::SameLine();
	if (ImGui::Button("Opal")) {
		Mix_ADLMIDI_setEmulator(ADLMIDI_OPL3_EMU_OPAL);
	}
	ImGui::SameLine();
	if (ImGui::Button("Java")) {
		Mix_ADLMIDI_setEmulator(ADLMIDI_OPL3_EMU_JAVA);
	}
	#endif
	ImGui::End();
}

if (sysinfo) {
	ImGui::Begin("System Information", &sysinfo, ImGuiWindowFlags_AlwaysAutoResize);
	ImGui::BulletText("SDL Revision: %s", SDL_GetRevision());
	ImGui::BulletText("Dear ImGui Release: %s", IMGUI_VERSION);
	ImGui::BulletText("Platform: %s", SDL_GetPlatform());
	ImGui::BulletText("Video driver: %s", SDL_GetCurrentVideoDriver());
	ImGui::BulletText("Audio driver: %s", SDL_GetCurrentAudioDriver());
	//ImGui::Text("Mixer ver: %d", Mix_Linked_Version());
	ImGui::BulletText("System RAM: %iMiB", SDL_GetSystemRAM());
	if (truecolor)
	ImGui::BulletText("Truecolour 24-bit mode enabled");
	if (!truecolor)
	ImGui::Text("256 colour-paletted 8-bit graphics mode in use");
	//int w, h;
	//SDL_RenderGetLogicalSize(SDL_GetRenderer(g_display->window), &w, &h);
	//ImGui::BulletText("Renderer logical size %d by %d", w, h);
	//float sx, sy;
	//SDL_RenderGetScale(SDL_GetRenderer(g_display->window), &sx, &sy);
	//ImGui::BulletText("Renderer scale is %f by %f", sx, sy);
	ImGui::Spacing();
	ImGui::SeparatorText("Paths");
	ImGui::BulletText("Name is: %s", paths_getdmodname());
	ImGui::BulletText("Location is: %s", paths_getdmoddir());
	ImGui::BulletText("Fallback is: %s", paths_getfallbackdir());
	ImGui::BulletText("Ancillary data is: %s", paths_getpkgdatadir());
	//ImGui::Separator();
	ImGui::Spacing();
	if (ImGui::SmallButton("Okay"))
	sysinfo = false;

	ImGui::End();
}

if (metrics) {
	ImGui::Begin("Metrics Graphs", &metrics, ImGuiWindowFlags_AlwaysAutoResize);
	if (ImGui::TreeNode("Active sprites")) {

		ImGui::TreePop();
	}
	ImGui::End();
}

if (mapsquares) {
	ImGui::Begin("Map Screens", &mapsquares, ImGuiWindowFlags_AlwaysAutoResize);
	ImGui::SeparatorText("Current Screen Data");
	ImGui::InputInt("MIDI Number", &g_dmod.map.music[*pplayer_map]);
	ImGui::InputInt("Screen reference (don't change this)", &g_dmod.map.loc[*pplayer_map]);
	ImGui::RadioButton("Outdoors", &g_dmod.map.indoor[*pplayer_map], 0);
	ImGui::SameLine();
	ImGui::RadioButton("Indoors", &g_dmod.map.indoor[*pplayer_map], 1);
	ImGui::SameLine();
	if (ImGui::Button("Save to Dink.dat")) {
		g_dmod.map.save();
	}
	ImGui::Separator();
	if (ImGui::BeginTable("table3", 32, ImGuiTableFlags_Borders))
{
    for (int item = 1; item < 769; item++)
    {
        ImGui::TableNextColumn();
		if (g_dmod.map.loc[item] > 0)
		ImGui::TableSetBgColor(ImGuiTableBgTarget_CellBg, ImGui::GetColorU32(ImVec4(1.0f, 0.1f, 0.2f, 0.65f)));
		else {
			ImGui::TableSetBgColor(ImGuiTableBgTarget_CellBg, ImGui::GetColorU32(ImVec4(0.0f, 0.1f, 1.0f, 0.65f)));
		}
		if (item == *pplayer_map)
		ImGui::TextColored(ImVec4(0,1,0,1), "%i", item);
		else if (g_dmod.map.loc[item] == 0) {

			ImGui::PushID(item * 100000);
			if (ImGui::SmallButton("New")) {
				g_dmod.map.loc[item] = add_new_map();
				g_dmod.map.save();
			}
			ImGui::PopID();
		}
		else {
			ImGui::Text("%i", item);
		}
		if (ImGui::IsItemHovered())
                ImGui::SetTooltip("Screen: %i (%i), Indoor: %i, Music: %i", item, g_dmod.map.loc[item],  g_dmod.map.indoor[item], g_dmod.map.music[item]);
    }
    ImGui::EndTable();
}
ImGui::End();
}

if (maptiles) {
	ImGui::Begin("Screen tiles", &maptiles, ImGuiWindowFlags_AlwaysAutoResize);
	//if (ImGui::SliderInt("Screen", &ts, 1, 41)) {
		//refresh_map_tiles();
	//}
	if (ImGui::Button("Reset all tiles")) {
		for (int i=0; i<96; i++)
            cur_ed_screen.t[i].square_full_idx0 = 0;
	}
	ImGui::SameLine();
	if (ImGui::Button("Reset all alt-hard")) {
        for (int i=0; i<96; i++)
            cur_ed_screen.t[i].althard = 0;
    }
	ImGui::Separator();
	ImGui::Text("Visible tile assignation and alt-hard");
    if (ImGui::BeginTable("table15", 12, ImGuiTableFlags_Borders)) {
        for (int i=0; i < 96; i++) {
            ImGui::TableNextColumn();
			ImGui::PushItemWidth(50.0f);
			ImGui::PushID(i * 4);
			ImGui::DragScalar("##sneed", ImGuiDataType_S16, &cur_ed_screen.t[i].square_full_idx0, 0.2f, &zero, &tilemax);
            //ImGui::TextColored(ImVec4(1.0f, 0.0f, 1.0f, 1.0f), "Index: %d", cur_ed_screen.t[i].square_full_idx0);
			ImGui::PopID();
			ImGui::PushID(i * 400);
			ImGui::DragScalar("##chuck", ImGuiDataType_S16, &cur_ed_screen.t[i].althard, 0.2f, &zero, &htilemax);
            //ImGui::TextColored(ImVec4(1.0f, 1.0f, 0.0f, 1.0f), "Alt: %d", cur_ed_screen.t[i].althard);
			ImGui::PopID();
			ImGui::PopItemWidth();
        }
		ImGui::EndTable();
	}
	if (ImGui::Button("Save changes to MAP.DAT")) {
			save_screen(g_dmod.map.map_dat.c_str(), g_dmod.map.loc[*pplayer_map]);
		}
	ImGui::End();
}

if (mapsprites) {
	ImGui::Begin("Sprites", &mapsprites);
		int seq_no = 75;
		SDL_Texture *texture = SDL_CreateTextureFromSurface(SDL_GetRenderer(g_display->window), GFX_k[seq[seq_no].frame[1]].k->screenshot());
		ImGui::Image(texture, ImVec2((int)GFX_k[seq[seq_no].frame[1]].k->w, GFX_k[seq[seq_no].frame[1]].k->h));
	ImGui::End();
}

if (dinkini) {
	ImGui::Begin("Dink.ini", &dinkini);
	ImGui::Checkbox("Show unused sequence slots", &freeslots);
	ImGui::Separator();
	if (ImGui::BeginTable("table9", 3, ImGuiTableFlags_Borders)) {
		ImGui::TableSetupColumn("Number", ImGuiTableColumnFlags_WidthFixed, 35.1f);
		ImGui::TableSetupColumn("ini line or init() command");
		ImGui::TableSetupColumn("Frames", ImGuiTableColumnFlags_WidthFixed, 35.1f);
		ImGui::TableHeadersRow();
		for (int i = 1; i < MAX_SEQUENCES; i++) {
			if ((!seq[i].is_active && freeslots) || seq[i].is_active) {
				ImGui::TableNextColumn();
				ImGui::Text("%i", i);
				ImGui::TableNextColumn();
				if (seq[i].ini != NULL)
				ImGui::Text("%s", seq[i].ini);
				else
				ImGui::TextColored(ImVec4(0.0f, 1.0f, 0.0f, 1.0f), "Empty");
				ImGui::TableNextColumn();
				ImGui::Text("%hi", seq[i].len);
			}

		}
		ImGui::EndTable();
	}
	ImGui::End();
}

if (htiles)
{
	ImGui::Begin("Hard.dat Tile Viewer", &htiles, ImGuiWindowFlags_AlwaysAutoResize);
	if (ImGui::BeginTabBar("hdat"))
	{
		if (ImGui::BeginTabItem("Hard.dat squares"))
		{
			ImVec4 standard = ImVec4(0.97f, 0.86f, 0.38f, 1.0f);
			ImVec4 low = ImVec4(0.3f, 0.4f, 1.0f, 1.0f);
			ImVec4 fire = ImVec4(1.0f, 0.0f, 0.2f, 1.0f);
			ImVec4 col;
			ImGui::SliderInt("Tile", &htile, 0, 799);
			ImGui::BeginTable("table9", 50);
			int y = 0;
			for (y = 0; y < 50; y++)
			{
				int x = 0;
				for (x = 0; x < 50; x++)
				{
				if (hmap.htile[htile].hm[x][y] == 1)
					col = standard;
				else if (hmap.htile[htile].hm[x][y] == 2)
					col = low;
				else if (hmap.htile[htile].hm[x][y] == 3)
					col = fire;
				if (hmap.htile[htile].hm[x][y] > 0)
		ImGui::TextColored(col, "%d", hmap.htile[htile].hm[x][y]);
				else
		ImGui::Text("0");
				ImGui::TableNextColumn();
				}
				ImGui::TableNextRow();
			}
			ImGui::EndTable();
			ImGui::EndTabItem();
		}

		if (ImGui::BeginTabItem("Tile Index"))
		{
			ImGui::SetNextItemWidth(50);
			ImGui::DragInt("Select Tilescreen", &tindexscr, 0.2f, 1, 41);
			ImGui::TextColored(ImVec4(1, 0.5, 0, 1), "Warning: Will save to HARD.DAT");
			if (ImGui::Button("Clear indices from tilescreen"))
			{
				for (int i = 128 * (tindexscr - 1); i < 128 * tindexscr; i++)
				hmap.btile_default[i] = 0;
				save_hard();
			}
			if (ImGui::IsItemHovered())
				ImGui::SetTooltip("Be careful!");
			if (ImGui::Button("Reset bogus index values to zero"))
			{
				for (int i = 0; i < 5248; i++)
				{
				if (hmap.btile_default[i] < 0 || hmap.btile_default[i] > 800)
					hmap.btile_default[i] = 0;
				}
				save_hard();
			}
			if (ImGui::BeginTable("table9", 2, ImGuiTableRowFlags_Headers))
			{
				ImGui::TableSetupColumn("Tile");
				ImGui::TableSetupColumn("Index");
				ImGui::TableHeadersRow();
				for (int i = (tindexscr - 1) * 128; i < ((tindexscr - 1) * 128) + 128; i++)
				{
				ImGui::TableNextColumn();
				ImGui::Text("%d", i);
				ImGui::TableNextColumn();
				ImGui::Text("%hd", hmap.btile_default[i]);
				}
				ImGui::EndTable();
			}
			ImGui::EndTabItem();
		}

		ImGui::EndTabBar();
	}
	ImGui::End();
}

if (magicitems) {
	ImGui::Begin("Magic and Items", &magicitems, ImGuiWindowFlags_AlwaysAutoResize);
	ImGui::Text("Green in the index column indicates the slot is in use");
	if (ImGui::BeginTabBar("magit")) {
	if (ImGui::BeginTabItem("Magic")) {
		if (ImGui::Button("Kill current magic")) {
			kill_cur_magic();
			draw_status_all();
		}
		ImGui::Separator();
		//Table
		if (ImGui::BeginTable("magictable", 4, ImGuiTableRowFlags_Headers)) {
			ImGui::TableSetupColumn("Index");
			ImGui::TableSetupColumn("Name");
			ImGui::TableSetupColumn("Sequence");
			ImGui::TableSetupColumn("Frame");
			ImGui::TableHeadersRow();

			for (int i = 0; i < NB_MITEMS; i++) {
				ImGui::TableNextColumn();
				if (play.mitem[i].active)
				ImGui::TextColored(ImVec4(0.0f, 1.0f, 0.0f, 1.0f), "%d", i+1);
				else
				ImGui::Text("%d", i+1);
				ImGui::TableNextColumn();
				//make button for arming in used slots
				if (play.mitem[i].active) {
				if (ImGui::Button(play.mitem[i].name)) {
					if (*pcur_magic != 0) {
					//disarm old weapon
					scripting_run_proc(magic_script, "DISARM");
				}
				*pcur_magic = i + 1;
				magic_script =
						scripting_load_script(play.mitem[*pcur_magic - 1].name, 1000, 1);
				scripting_run_proc(magic_script, "ARM");
				scripting_run_proc(magic_script, "ARMMOVIE");
				draw_status_all();
				}
				}
				else
				ImGui::Text("%s", play.mitem[i].name);
				ImGui::TableNextColumn();
				ImGui::Text("%d", play.mitem[i].seq);
				ImGui::TableNextColumn();
				ImGui::Text("%d", play.mitem[i].frame);
		}
		ImGui::EndTable();
		}
		ImGui::EndTabItem();
	}

	//Items
	if (ImGui::BeginTabItem("Items and/or Weapons")) {
		if (ImGui::Button("Kill current item/weapon")) {
			kill_cur_item();
			draw_status_all();
		}
		ImGui::Separator();
		//Table
		if (ImGui::BeginTable("itable", 4, ImGuiTableRowFlags_Headers)) {
			ImGui::TableSetupColumn("Index");
			ImGui::TableSetupColumn("Name");
			ImGui::TableSetupColumn("Sequence");
			ImGui::TableSetupColumn("Frame");
			ImGui::TableHeadersRow();

			for (int i = 0; i < NB_ITEMS; i++) {
				ImGui::TableNextColumn();
				if (play.item[i].active)
				ImGui::TextColored(ImVec4(0.0f, 1.0f, 0.0f, 1.0f), "%d", i+1);
				else
				ImGui::Text("%d", i+1);
				ImGui::TableNextColumn();
				if (play.item[i].active) {
					if (ImGui::Button(play.item[i].name)) {
						if (*pcur_weapon != 0) {
					//disarm old weapon
					scripting_run_proc(weapon_script, "DISARM");
				}

				*pcur_weapon = i + 1;
				weapon_script =
						scripting_load_script(play.item[*pcur_weapon - 1].name, 1000, 1);
				scripting_run_proc(magic_script, "ARM");
				scripting_run_proc(magic_script, "ARMMOVIE");
				draw_status_all();
					}
				}
				else
				ImGui::Text("%s", play.item[i].name);
				ImGui::TableNextColumn();
				ImGui::Text("%d", play.item[i].seq);
				ImGui::TableNextColumn();
				ImGui::Text("%d", play.item[i].frame);

			}
			ImGui::EndTable();
		}
		ImGui::EndTabItem();
	}

	ImGui::EndTabBar();
	}

	ImGui::End();
}

if (helpbox) {
	ImGui::Begin("A Brief Description of YeOldeDink", &helpbox);
	ImGui::TextWrapped("Welcome to YeOldeDink, a fork of Ghostknight's GNU Freedink fork with several bugfixes and enhancements. As a general reminder, most input boxes can be typed into by activating the console or by mousing over the relevant window, and slider/drag widgets can have their increment speed modified by holding either shift or alt while interacting.\n\n");
	ImGui::Spacing();
	ImGui::SeparatorText("File menu");
	ImGui::BulletText("Hide menubar and debug interface: This closes the ImGui interface and is equivalent to pressing Alt+D\n");
	ImGui::BulletText("Pause: This interrupts the execution of the engine until you choose to unpause\n");
	ImGui::BulletText("Restart: This will take you back to the title screen or similar\n");
	ImGui::BulletText("Quit: This will terminate the process and return you to your OS environment\n\n");
	ImGui::Spacing();
	ImGui::SeparatorText("Debug menu");
	ImGui::BulletText("Dev mode: switching this on makes visible many of the settings listed below.");
	ImGui::TextWrapped("- Info overlay: A transparent window that shows your FPS, along with the last debug log entry. The window may be moved to any corner of the screen or to a custom location by right-clicking. The DinkC console will appear here when activated.\n");
	ImGui::TextWrapped("- Game popout window: Dink within Dink. Useful if you want to go full-screen while using the debug log at the same time etc\n");
	ImGui::TextWrapped("- Ultimat(er) Cheat: Homage to the much-beloved script by Gary Hertel and Ted Shutes.\n");
	ImGui::TextWrapped("- Debug Log: See engine messages in real-time. Log level selects which messages are displayed, and some of the more useless log triggers have been switched off by default. The DinkC console will also appear in this window when enabled.\n");
	ImGui::TextWrapped("- DinkC Console: manually toggles the console. Equivalent to pressing Alt+C. It will also bypass keyboard input control for typing into text boxes etc\n");
	ImGui::TextWrapped("- Pause while console activated: The execution of the engine will be interrupted, and your DinkC command results apparent after deactivating the console while enabled\n");
	ImGui::TextWrapped("- Render game in main window: Toggles the display to the background. Designed for use with the popout window.\n");
	ImGui::TextWrapped("- Pink squares: Shows a pink hardbox overlay on touchable sprites, as well as bounding boxes for all sprites upon punching things.\n");
	ImGui::TextWrapped("- hard.dat overlay: Display tile and sprite collision for the screen without having to open DinkEdit.");
	//ImGui::TextWrapped("- Render tiles and background: Turns off the re-rendering of the background upon a screen change. Probably best to leave it on.\n");
	ImGui::TextWrapped("- Draw invisible sprites: Draws sprites that are either set to type 2, or are specified nodraw.\n");
	ImGui::TextWrapped("- Ignore sprite visions: Display all sprites on a screen regardless of which vision they're assigned.\n");
	//ImGui::TextWrapped("- Write to debug.txt: Chooses if messages are appended to debug.txt\n\n");
	ImGui::Spacing();
	ImGui::SeparatorText("Display menu");
	ImGui::BulletText("Texture filtering: Choose whether to filter the screen texture or use nearest-neighbour sampling");
	ImGui::BulletText("Ignore aspect ratio: fills the entire screen rather than using 4:3");
	ImGui::BulletText("Screen transition: Allows you to switch off the slide effect upon changing screens");
	ImGui::BulletText("Smooth shadows: A half-baked reimplementation of Redink1's Good Shadows. Doesn't work on Fastfiles.");
	ImGui::BulletText("Smooth fade: uses tweening algorithms to set the brightness during fades to look nicer");
	ImGui::BulletText("Set window to...: Changes the resolution in windowed mode.\n");
	ImGui::BulletText("Toggle full-screen display: Equivalent to pressing Alt+Enter.\n");
	ImGui::BulletText("Make window borderless: Removes the menu bar and outline provided by the operating system.");
	ImGui::Spacing();
	ImGui::SeparatorText("System menu");
	ImGui::TextWrapped("- Gamepad and Joystick: See whether or not your controller is detected, and attempt to connect it if it isn't.\n");
	ImGui::TextWrapped("- Audio: Adjusts a variety of sound-related settings such as volume and MIDI playback.\n");
	ImGui::TextWrapped("- System information: Shows your linked SDL2 version along with the video/audio backend in use. Also reminds you of how much RAM you have, and your operating system in case you forgot.\n\n");
	ImGui::Spacing();
	ImGui::SeparatorText("Engine data menu");
	ImGui::TextWrapped("- Sprites: On the first tab, shows you active sprites on the current screen along with some parameters. The second tab will allow you to edit these parameters in real-time, and the third tab edits the sprite properties in MAP.DAT itself");
	ImGui::TextWrapped("- Variables: View and edit your local and global variables. Filter out globals with the checkbox.\n");
	ImGui::TextWrapped("- Scripts: Lists which scripts are running at the moment, along with which part of the file is being read.\n");
	ImGui::TextWrapped("- Editor Overrides: allows you to view and edit editor_frame/type/seq values on the current screen");
	ImGui::TextWrapped("- Map Index: An overview of which screens are in use, which screen you're on, with the ability to create new screens or edit MIDI/indoor. Mouse over to see the dink.dat screen number and the corresponding parameters");
	ImGui::TextWrapped("- Dink.ini: See which sequences are in use, and which are empty. Note that the list may not be entirely accurate due to SET_FRAME_FRAME commands etc.");
	ImGui::TextWrapped("- Colour palette: see the current palette in use and select a colour to fill the screen with. You may also edit individual colours on the second tab.\n\n");
	ImGui::Spacing();
	ImGui::SeparatorText("Speed");
	ImGui::BulletText("Normal speed: Equivalent to releasing the Tab key outside of debug mode after having held it down.\n");
	ImGui::BulletText("High speed: This will run the engine at twice the usual speed until chosen otherwise.");
	ImGui::BulletText("Slow speed: This runs the engine at half the normal speed. Unavailable outside of debug mode.");
	ImGui::Spacing();
	ImGui::SeparatorText("Quick save/load");
	ImGui::BulletText("QuickSave #Whatever: Saves the game into very high-numbered slots");
	ImGui::BulletText("Load: Reads a save file from a corresponding slot. Not available in mouse mode.\n");
	ImGui::Spacing();
	ImGui::SeparatorText("Other stuff");
	ImGui::BulletText("PNG support for both sprites and tiles in 24-bit mode");
	ImGui::BulletText("MP3 support for 1.08 Aural+ enjoyers, as well as FLAC and Opus (and SPC/VGZ) with playmidi()");
	ImGui::BulletText("MP3, Ogg, and FLAC sound effects with playsound()");
	//ImGui::BulletText("44.1kHz audio playback!");
	ImGui::BulletText("Pressing Tab in mouse mode will centre the cursor on the screen");
	ImGui::BulletText("Alt+U will unfreeze sprite 1 (player/mouse)");
	ImGui::BulletText("Controller disconnects will be autodetected, as will reconnects");
	ImGui::BulletText("DinkC: get_client_fork() returns 3");
	ImGui::BulletText("DinkC: get_platform() will return a value consistent with RTDink");
	ImGui::BulletText("DinkC: play_mod_order(int) will set the pattern for tracker music");
	ImGui::BulletText("DinkC: playavi(video.flv) will play back a video using ffplay. Requires FFmpeg to be installed.");
	ImGui::End();
}

if (aboutbox) {
	ImGui::Begin("About", &aboutbox, ImGuiWindowFlags_AlwaysAutoResize);
	ImGui::InvisibleButton("sneed", ImVec2(1.0f, 1.0f));
	ImGui::SameLine();
	ImGui::Text("YeOldeDink %s © 2022-2024 YeOldeToast", VERSION);
	ImGui::InvisibleButton("feed", ImVec2(1.0f, 1.0f));
	ImGui::SameLine();
	ImGui::Text("Based upon GNU Freedink by Beuc, Phoenix (Alexander Krivács Schrøder), and others, with improvements by Ghostknight");
	ImGui::InvisibleButton("chuck", ImVec2(1.0f, 1.0f));
	ImGui::SameLine();
	ImGui::Text("Many thanks to drone1400, TealCool, @MakhnoZhiv, and everyone else keeping Dink alive.");
	ImGui::Spacing();
	ImGui::TextWrapped("Uses Dear ImGui by Omar Cornut, Backward-cpp by bombela, Random for Modern C++ by effalkronium, Indicators and Glob by p-ranav, simpleini by brofield, Tweeny and Asap by Mobius3, Tinyfiledialogs by Guillaume Vareille, filesystem by gulrak, parallel-hashmap by greg7mdp, Lua5.4 and Cereal");
	ImGui::SeparatorText("Licenses");
	if (ImGui::TreeNode("GNU General Public License Version 3")) {
	ImGui::TextWrapped(
	"YeOldeDink is free software: you can redistribute it and/or modify "
    "it under the terms of the GNU General Public License as published by "
    "the Free Software Foundation, either version 3 of the License, or "
    "(at your option) any later version.");
	ImGui::Spacing();
	ImGui::TextWrapped(
		"YeOldeDink is distributed in the hope that it will be useful, "
    "but WITHOUT ANY WARRANTY; without even the implied warranty of "
    "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the "
    "GNU General Public License for more details.");
	ImGui::Spacing();
	ImGui::TextWrapped(
		"You should have received a copy of the GNU General Public License "
		"along with this program. If not, see <http://www.gnu.org/licenses/>.");
		ImGui::TreePop();
	}
	if (ImGui::TreeNode("MIT License")) {
	ImGui::TextWrapped(
"Copyright (c) 2014-2024 Omar Cornut, Copyright (c) 1994–2019 Lua.org, PUC-Rio, Copyright (c) 2020-2022 Eduardo Bart (https://github.com/edubart), Copyright (c) 2019 Pranav, © 2017-2023 effolkronium (Illia Polishchuk), (c) 2006-2022 Brodie Thiesfield, 2013 Google Inc. All Rights Reserved.\n\n"
"Permission is hereby granted, free of charge, to any person obtaining a copy "
"of this software and associated documentation files (the \"Software\"), to deal "
"in the Software without restriction, including without limitation the rights "
"to use, copy, modify, merge, publish, distribute, sublicense, and/or sell "
" copies of the Software, and to permit persons to whom the Software is "
"furnished to do so, subject to the following conditions:\n\n "
"The above copyright notice and this permission notice shall be included in all "
"copies or substantial portions of the Software.\n\n"
"THE SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR "
"IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, "
"FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE "
"AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER "
"LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, "
"OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE "
"SOFTWARE.");
ImGui::TreePop();
	}
//ImGui::Spacing();
if (ImGui::TreeNode("BSD")) {
ImGui::TextWrapped(
	"SoLoud audio engine \n"
"Copyright (c) 2013-2018 Jari Komppa \n"
"Copyright (c) 2013-2022, Randolph Voorhies, Shane Grant\n\n"

"This software is provided 'as-is', without any express or implied "
"warranty. In no event will the authors be held liable for any damages "
"arising from the use of this software. \n\n"

"Permission is granted to anyone to use this software for any purpose, "
"including commercial applications, and to alter it and redistribute it "
"freely, subject to the following restrictions: \n\n"

"   1. The origin of this software must not be misrepresented; you must not  "
"claim that you wrote the original software. If you use this software "
"in a product, an acknowledgment in the product documentation would be "
"appreciated but is not required.\n\n"

"   2. Altered source versions must be plainly marked as such, and must not be "
"   misrepresented as being the original software. \n\n"

"   3. This notice may not be removed or altered from any source"
" distribution."
);
ImGui::TreePop();
}
if (ImGui::TreeNode("Apache 2.0")) {
	ImGui::TextWrapped(
	"Licensed under the Apache License, Version 2.0 (the \"License\");"
   "you may not use this file except in compliance with the License."
  " You may obtain a copy of the License at\n\n"
      " https://www.apache.org/licenses/LICENSE-2.0\n\n"
   "Unless required by applicable law or agreed to in writing, software"
   "distributed under the License is distributed on an \"AS IS\" BASIS,"
   "WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied."
  " See the License for the specific language governing permissions and"
   "limitations under the License."
	);

	ImGui::TreePop();
}
/*
if (ImGui::TreeNode("SIL OFL")) {
	ImGui::TextWrapped(
"-----------------------------------------------------------
SIL OPEN FONT LICENSE Version 1.1 - 26 February 2007
-----------------------------------------------------------

PREAMBLE
The goals of the Open Font License (OFL) are to stimulate worldwide
development of collaborative font projects, to support the font creation
efforts of academic and linguistic communities, and to provide a free and
open framework in which fonts may be shared and improved in partnership
with others.

The OFL allows the licensed fonts to be used, studied, modified and
redistributed freely as long as they are not sold by themselves. The
fonts, including any derivative works, can be bundled, embedded,
redistributed and/or sold with any software provided that any reserved
names are not used by derivative works. The fonts and derivatives,
however, cannot be released under any other type of license. The
requirement for fonts to remain under this license does not apply
to any document created using the fonts or their derivatives.

DEFINITIONS
\"Font Software\" refers to the set of files released by the Copyright
Holder(s) under this license and clearly marked as such. This may
include source files, build scripts and documentation.

\"Reserved Font Name\" refers to any names specified as such after the
copyright statement(s).

\"Original Version\" refers to the collection of Font Software components as
distributed by the Copyright Holder(s).

\"Modified Version\" refers to any derivative made by adding to, deleting,
or substituting -- in part or in whole -- any of the components of the
Original Version, by changing formats or by porting the Font Software to a
new environment.

\"Author\" refers to any designer, engineer, programmer, technical
writer or other person who contributed to the Font Software.

PERMISSION & CONDITIONS
Permission is hereby granted, free of charge, to any person obtaining
a copy of the Font Software, to use, study, copy, merge, embed, modify,
redistribute, and sell modified and unmodified copies of the Font
Software, subject to the following conditions:

1) Neither the Font Software nor any of its individual components,
in Original or Modified Versions, may be sold by itself.

2) Original or Modified Versions of the Font Software may be bundled,
redistributed and/or sold with any software, provided that each copy
contains the above copyright notice and this license. These can be
included either as stand-alone text files, human-readable headers or
in the appropriate machine-readable metadata fields within text or
binary files as long as those fields can be easily viewed by the user.

3) No Modified Version of the Font Software may use the Reserved Font
Name(s) unless explicit written permission is granted by the corresponding
Copyright Holder. This restriction only applies to the primary font name as
presented to the users.

4) The name(s) of the Copyright Holder(s) or the Author(s) of the Font
Software shall not be used to promote, endorse or advertise any
Modified Version, except to acknowledge the contribution(s) of the
Copyright Holder(s) and the Author(s) or with their explicit written
permission.

5) The Font Software, modified or unmodified, in part or in whole,
must be distributed entirely under this license, and must not be
distributed under any other license. The requirement for fonts to
remain under this license does not apply to any document created
using the Font Software.

TERMINATION
This license becomes null and void if any of the above conditions are
not met.

DISCLAIMER
THE FONT SOFTWARE IS PROVIDED \"AS IS\", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO ANY WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT
OF COPYRIGHT, PATENT, TRADEMARK, OR OTHER RIGHT. IN NO EVENT SHALL THE
COPYRIGHT HOLDER BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
INCLUDING ANY GENERAL, SPECIAL, INDIRECT, INCIDENTAL, OR CONSEQUENTIAL
DAMAGES, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF THE USE OR INABILITY TO USE THE FONT SOFTWARE OR FROM
OTHER DEALINGS IN THE FONT SOFTWARE.
");
ImGui::TreePop();
}
*/
	if (ImGui::SmallButton("Okay")) {
	aboutbox = false;
	play_sneedle();
	}
	ImGui::End();
}

if (luaviewer) {
	ImGui::Begin("DinkLua Variables", &luaviewer);


	ImGui::End();
}

if (sproverride) {
	ImGui::Begin("Editor sprite overrides and save data", &sproverride);
	time_t ct;
	time(&ct);
	if (mode < 3)
	ImGui::BeginDisabled();
	ImGui::BulletText("Game timer: %f min", play.minutes + (difftime(ct, time_start) / 60));
	ImGui::SameLine();
	if (ImGui::SmallButton("Reset"))
	play.minutes = 0;
	ImGui::BulletText("Screen timer: %u seconds / %u min", play.spmap[*pplayer_map].last_time / 1000, (play.spmap[*pplayer_map].last_time / 1000) / 60);
	ImGui::SameLine();
	if (ImGui::SmallButton("Reset"))
	play.spmap[*pplayer_map].last_time = 0;
	ImGui::InputText("Savegame Info", save_game_info, sizeof(save_game_info));
	ImGui::Spacing();
	ImGui::SeparatorText("Editor Sprite Overrides");
	ImGui::Text("Selected sprite: %d", sproversel);
	ImGui::SameLine();
	if (ImGui::Button("Reset this sprite's override values to zero")) {
		play.spmap[*pplayer_map].type[sproversel] = 0;
		play.spmap[*pplayer_map].frame[sproversel] = 0;
		play.spmap[*pplayer_map].seq[sproversel] = 0;
	}

	ImGui::PushItemWidth(30.0f);
	ImGui::InputScalar("Type", ImGuiDataType_U8, &play.spmap[*pplayer_map].type[sproversel]);
	ImGui::InputScalar("Frame", ImGuiDataType_U8, &play.spmap[*pplayer_map].frame[sproversel]);
	ImGui::InputScalar("Sequence", ImGuiDataType_U8, &play.spmap[*pplayer_map].seq[sproversel]);
	ImGui::PopItemWidth();
	ImGui::Separator();
	ImGui::Checkbox("Show only modified", &sprovermod);
	ImGui::SameLine();
		if (ImGui::Button("Reset all values to zero")) {
		for (int j=1; j<100; j++) {
			play.spmap[*pplayer_map].type[j] = 0;
			play.spmap[*pplayer_map].seq[j] = 0;
			play.spmap[*pplayer_map].frame[j] = 0;
		}
	}
	if (ImGui::BeginTable("table88", 4, ImGuiTableFlags_Borders)) {
		ImGui::TableSetupColumn("Sprite");
		ImGui::TableSetupColumn("Type");
		ImGui::TableSetupColumn("Frame");
		ImGui::TableSetupColumn("Sequence");
		ImGui::TableHeadersRow();
		for (int i=1; i<100; i++) {
			if ((sprovermod && play.spmap[*pplayer_map].type[i] != 0) || !sprovermod) {
			ImGui::TableNextColumn();
			ImGui::PushID(i);
			if (ImGui::Button("Select")) {
				sproversel = i;
			}
			ImGui::PopID();
			ImGui::SameLine();
			ImGui::Text("%d", i);
			ImGui::TableNextColumn();
			if (play.spmap[*pplayer_map].type[i] == 1)
			ImGui::Text("1: Killed completely");
			else if (play.spmap[*pplayer_map].type[i] == 2)
			ImGui::Text("2: Drawn nohard");
			else if (play.spmap[*pplayer_map].type[i] == 3)
			ImGui::Text("3: Background nohard");
			else if (play.spmap[*pplayer_map].type[i] == 4)
			ImGui::Text("4: Drawn hard");
			else if (play.spmap[*pplayer_map].type[i] == 5)
			ImGui::Text("5: Background hard");
			else if (play.spmap[*pplayer_map].type[i] == 6)
			ImGui::Text("6: Returns in 5 minutes");
			else if (play.spmap[*pplayer_map].type[i] == 7)
			ImGui::Text("7: Returns in 3 minutes");
			else if (play.spmap[*pplayer_map].type[i] == 8)
			ImGui::Text("8: Returns in 1 minute");
			else
			ImGui::Text("%d", play.spmap[*pplayer_map].type[i]);
			ImGui::TableNextColumn();
			ImGui::Text("%d", play.spmap[*pplayer_map].frame[i]);
			ImGui::TableNextColumn();
			ImGui::Text("%d", play.spmap[*pplayer_map].seq[i]);
			if (i == sproversel)
			ImGui::TableSetBgColor(ImGuiTableBgTarget_RowBg0, ImGui::GetColorU32(ImVec4(0.7f, 0.3f, 0.3f, 0.65f)));
			}
		}

		ImGui::EndTable();
	}
	if (mode < 3)
	ImGui::EndDisabled();
	ImGui::End();
}

if (paledit) {
	ImGui::Begin("Colour Palette", &paledit, ImGuiWindowFlags_AlwaysAutoResize);
	if (ImGui::BeginTabBar("colpal")) {
	if (ImGui::BeginTabItem("View")) {
		if (ImGui::InputText("Load palette from BMP", bmppal, IM_ARRAYSIZE(bmppal), ImGuiInputTextFlags_EnterReturnsTrue)) {
			gfx_palette_set_from_bmp(bmppal);
		}
		ImGui::SameLine();
		if (ImGui::SmallButton("Set")) {
			gfx_palette_set_from_bmp(bmppal);
			memset(bmppal, 0, sizeof(bmppal));
		}
	ImGui::Text("Click on a colour to fill the background layer with it, equivalent to fill_screen()");
	SDL_Color palette[256];
	gfx_palette_get_phys(palette);
	//ImVec4 col = {palette[4].r / 255.0f, (float)palette[4].g / 255.0f, (float)palette[4].b / 255.0f, 1.0f};
	//ImGui::ColorPicker4("Edit", (float*)&col);
	for (int i=0; i< 256; i++) {
		ImGui::PushID(i * 1000000);
		if (ImGui::ColorButton("##col", ImVec4((float)palette[i].r / 255.0f, (float)palette[i].g / 255.0f, (float)palette[i].b / 255.0f, 1.0f), ImGuiColorEditFlags_NoAlpha)) {
		IOGFX_background->fill_screen(i, GFX_ref_pal);
		palcol = i;
		}
		ImGui::PopID();
		if ((i == 0) || ((i+1) % 32 != 0))
		ImGui::SameLine();
	}
	ImGui::Separator();
	ImGui::Text("Palette index number: %d", palcol);
	ImGui::SameLine();
	ImGui::ColorButton("Selected Palette Colour",ImVec4((float)palette[palcol].r / 255.0f, (float)palette[palcol].g / 255.0f, (float)palette[palcol].b / 255.0f, 1.0f));
	if (ImGui::Button("Copy fill_screen() command to clipboard")) {
		ImGui::LogToClipboard();
		ImGui::LogText("fill_screen(%d)", palcol);
		ImGui::LogFinish();
	}
	ImGui::SameLine();
	if (ImGui::Button("Reset background and status bar")) {
	draw_screen_game_background();
	draw_status_all();
	}
	ImGui::SameLine();
	if (ImGui::Button("Reset palette")) {
		//gfx_palette_reset();
		gfx_palette_set_from_bmp("TILES/SPLASH.BMP");
	}
	ImGui::EndTabItem();
	}

	if (ImGui::BeginTabItem("Edit")) {
		if (ImGui::Button("Copy from screen palette")) {
			memcpy(mypal, GFX_ref_pal, sizeof(mypal));
		}
		for (int i=0; i< 256; i++) {
			//This might conflict with something
		ImGui::PushID(i * 1000);
		if (ImGui::ColorButton("##col", ImVec4((float)mypal[i].r / 255.0f, (float)mypal[i].g / 255.0f, (float)mypal[i].b / 255.0f, 1.0f), ImGuiColorEditFlags_NoLabel | ImGuiColorEditFlags_NoAlpha | ImGuiColorEditFlags_NoLabel)) {
		//IOGFX_background->fill_screen(i, mypal);
		mycol = i;
		mypalcol = ImVec4((float)mypal[mycol].r / 255.0f, (float)mypal[mycol].g / 255.0f, (float)mypal[mycol].b / 255.0f, (float)mypal[mycol].a / 255.0f);
		}
		ImGui::PopID();
		if ((i == 0) || ((i+1) % 32 != 0))
		ImGui::SameLine();
	}
	//ImGui::Separator();
	//ImGui::ColorButton("Selected Palette Colour",ImVec4((float)mypal[mycol].r / 255.0f, (float)mypal[mycol].g / 255.0f, (float)mypal[mycol].b / 255.0f, (float)mypal[mycol].a / 255.0f));
	ImGui::Spacing();
	ImGui::SeparatorText("Palette operations");
	//ImGui::Checkbox("Auto-apply", &autopal);
	if (ImGui::Button("Convert to greyscale")) {
		for (int c = 0; c < 256; c++) {
			color::rgb<std::uint8_t> c1( {mypal[c].r, mypal[c].g, mypal[c].b} );
			color::gray<std::uint8_t> c2;
			c2 = c1;
			mypal[c].r = ::color::get::red(c2);
			mypal[c].g = ::color::get::green(c2);
			mypal[c].b = ::color::get::blue(c2);
	}
	}
	ImGui::SameLine();
	if (ImGui::Button("Invert")) {
		for (int c = 0; c < 256; c++) {
			color::rgb<std::uint8_t> c1( {mypal[c].r, mypal[c].g, mypal[c].b} );
			::color::operation::invert(c1);
			mypal[c].r = ::color::get::red(c1);
			mypal[c].g = ::color::get::green(c1);
			mypal[c].b = ::color::get::blue(c1);
		}
	}
	ImGui::Spacing();
	ImGui::Text("HSV/L/I of all colours");
	ImGui::Text("H: ");
	ImGui::SameLine();
	if (ImGui::ArrowButton("hue-", ImGuiDir_Left)) {
		changehsv(0, -1, 0);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	if (ImGui::ArrowButton("hue+", ImGuiDir_Right)) {
		changehsv(0, 1, 0);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	ImGui::Text("S: ");
	ImGui::SameLine();
	if (ImGui::ArrowButton("S-", ImGuiDir_Left)) {
		changehsv(1, -1, 0);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	if (ImGui::ArrowButton("S+", ImGuiDir_Right)) {
		changehsv(1, 1, 0);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	ImGui::Text("V: ");
	ImGui::SameLine();
	if (ImGui::ArrowButton("V-", ImGuiDir_Left)) {
		changehsv(2, -1, 0);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	if (ImGui::ArrowButton("V+", ImGuiDir_Right)) {
		changehsv(2, 1, 0);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	ImGui::Text("L: ");
	ImGui::SameLine();
	if (ImGui::ArrowButton("L-", ImGuiDir_Left)) {
		changehsv(2, -1, 1);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	if (ImGui::ArrowButton("L+", ImGuiDir_Right)) {
		changehsv(2, 1, 1);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	ImGui::Text("I: ");
	ImGui::SameLine();
	if (ImGui::ArrowButton("I-", ImGuiDir_Left)) {
		changehsv(2, -1, 2);
		gfx_palette_set_phys(mypal);
	}
	ImGui::SameLine();
	if (ImGui::ArrowButton("I+", ImGuiDir_Right)) {
		changehsv(2, 1, 2);
		gfx_palette_set_phys(mypal);
	}
	//Picker
	if (ImGui::TreeNode("Colour Picker")) {
		if (ImGui::ColorPicker4("Edit", (float*)&mypalcol)) {
			mypal[mycol].r = mypalcol.x * 255;
			mypal[mycol].g = mypalcol.y * 255;
			mypal[mycol].b = mypalcol.z * 255;
			mypal[mycol].a = mypalcol.w * 255;
	}
	ImGui::TreePop();
	}
	if (ImGui::Button("Apply palette to screen")) {
		gfx_palette_set_phys(mypal);
		//mypalcol = ImVec4((float)mypal[mycol].r / 255.0f, (float)mypal[mycol].g / 255.0f, (float)mypal[mycol].b / 255.0f, (float)mypal[mycol].a / 255.0f);
	}
	ImGui::SameLine();
	if (ImGui::Button("Copy JASC PAL to clipboard")) {
		ImGui::LogToClipboard();
		ImGui::LogText("JASC-PAL\n");
		ImGui::LogText("0100\n");
		ImGui::LogText("256\n");
		for (int i = 0; i < 256; i++) {
			ImGui::LogText("%d %d %d\n", mypal[i].r, mypal[i].g, mypal[i].b);
		}
		ImGui::LogFinish();
	}

	ImGui::EndTabItem();
	}


	ImGui::EndTabBar();
	}
	ImGui::End();
}

if (iniconf) {
ImGui::Begin("yedink.ini file settings", &iniconf, ImGuiWindowFlags_AlwaysAutoResize);
if (ImGui::Button("Load config file data")) {
	CSimpleIniA ini;
	ini.SetUnicode();

	SI_Error rc = ini.LoadFile(paths_pkgdatafile("yedink.ini"));
	if (rc < 0) {
		log_warn("📋 Failed to open yedink.ini");
	} else {
		log_info("📋 Opened the ini file");
	}

	dispfont = ini.GetValue("fonts", "display");
	debugfont = ini.GetValue("fonts", "debugttf");
	strcpy(dispfon, dispfont);
	strcpy(debugfon, debugfont);
	debug_fontsize = atoi(ini.GetValue("fonts", "debug_pt_size", "172"));
	window_w = atoi(ini.GetValue("display", "window_w", "640"));
	window_h = atoi(ini.GetValue("display", "window_h", "480"));
	dispfilt = ini.GetBoolValue("display", "filtering", true);
	dispstretch = ini.GetBoolValue("display", "ignore_ratio", false);
	audio_samplerate = atoi(ini.GetValue("audio", "sameplerate", "48000"));
	audiobuf = atoi(ini.GetValue("audio", "buffer", "1024"));
	audiovol = atoi(ini.GetValue("audio", "music_volume", "128"));
	sfx_vol = atoi(ini.GetValue("audio", "sfx_volume", "128"));
	dbgtxtwrite = ini.GetBoolValue("debug", "write_debugtxt", false);
	debug_conpause = ini.GetBoolValue("debug", "console_pause", false);
}
ImGui::TextWrapped("These settings will only be apparent after saving and restarting");
ImGui::SeparatorText("Fonts");
ImGui::TextWrapped("The display font may either be a TTF file as part of resources, or something that Fontconfig can find. The debug interface font can only be a resources TTF file.");
ImGui::InputText("Display font", dispfon, IM_ARRAYSIZE(dispfon));
ImGui::InputText("Debug interface font", debugfon, IM_ARRAYSIZE(debugfon));
ImGui::SliderInt("Debug font size", &debug_fontsize, 8, 72);
ImGui::Spacing();
ImGui::SeparatorText("Display");
if (ImGui::SmallButton("640x480")) {
	window_w = 640;
	window_h = 480;
}
ImGui::SameLine();
if (ImGui::SmallButton("1280x960")) {
	window_w = 1280;
	window_h = 960;
}
ImGui::Text("CTRL+Click to type in your own values");
ImGui::DragInt("Window width", &window_w);
ImGui::DragInt("Window height", &window_h);
ImGui::Checkbox("Texture filtering", &dbg.debug_interp);
ImGui::Checkbox("Stretch beyond 4:3 aspect ratio", &dbg.debug_assratio);
ImGui::Spacing();
ImGui::SeparatorText("Audio");
if (ImGui::SmallButton("44.1Khz"))
audio_samplerate = 44100;
ImGui::SameLine();
if (ImGui::SmallButton("48kHz"))
audio_samplerate = 48000;
ImGui::SameLine();
if (ImGui::SmallButton("96kHz"))
audio_samplerate = 96000;
ImGui::Text("Set this to your sound card's actual rate for best performance");
ImGui::InputInt("Sampling rate", &audio_samplerate);
//ImGui::SliderInt("Buffer", &audiobuf, 32, 8192);
//ImGui::SliderInt("Music volume", &muvol, 1, 128);
//ImGui::SliderInt("SFX volume", &sfx_vol, 1, 128);
ImGui::Spacing();
ImGui::SeparatorText("Debug options");
ImGui::Checkbox("Show transparent info window", &dswitch);
//ImGui::Checkbox("Write to debug.txt", &dbg.debug_conwrite);
//ImGui::Checkbox("Pause while console active", &debug_conpause);
ImGui::Separator();

if (ImGui::Button("Save configuration to yedink.ini")) {
	CSimpleIniA ini;
	SI_Error rc = ini.LoadFile(paths_pkgdatafile("yedink.ini"));
	if (rc < 0) {
		log_warn("Failed to open yedink.ini");
	} else {
		log_info("Opened the ini file");
	}
	rc = ini.SetValue("fonts", "display", dispfon);
	rc = ini.SetValue("fonts", "debugttf", debugfon);
	char str[40];
	sprintf(str, "%d", debug_fontsize);
	rc = ini.SetValue("fonts", "debug_pt_size", str);
	sprintf(str, "%d", window_w);
	rc = ini.SetValue("display", "window_w", str);
	sprintf(str, "%d", window_h);
	rc = ini.SetValue("display", "window_h", str);
	rc = ini.SetBoolValue("display", "filtering", dbg.debug_interp);
	rc = ini.SetBoolValue("display", "ignore_ratio", debug_assratio);
	rc = ini.SetBoolValue("debug", "info_window", dswitch);
	rc = ini.SetBoolValue("debug", "write_debugtxt", dbg.debug_conwrite);
	sprintf(str, "%d", audio_samplerate);
	rc = ini.SetValue("audio", "samplerate", str);
	//sprintf(str, "%d", muvol);
	rc = ini.SetValue("audio", "music_volume", str);
	rc = ini.SaveFile(paths_pkgdatafile("yedink.ini"));
}
//if (ImGui::Button("Play video file")) {
//	unsigned int n = std::thread::hardware_concurrency();
//    log_info("%d Threads are supported", n);
//}
ImGui::End();
}

if (boxprev) {
	ImGui::Begin("Box previewer", &boxprev, ImGuiWindowFlags_AlwaysAutoResize);
	//ImGui::Checkbox("Preview bounds", &bprevview);
	ImGui::InputDouble("Left", &bprev.X.Min);
	ImGui::InputDouble("Top", &bprev.Y.Min);
	ImGui::InputDouble("Right", &bprev.X.Max);
	ImGui::InputDouble("Bottom", &bprev.Y.Max);
	if (ImPlot::BeginPlot("Box alignment", ImVec2(640, 480), ImPlotFlags_NoLegend)) {
		ImPlot::SetupAxesLimits(0, xmax, 0, ymax);
		ImPlot::SetupAxis(ImAxis_Y1, NULL, ImPlotAxisFlags_Invert | ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin | ImPlotAxisFlags_NoTickLabels);
		ImPlot::SetupAxis(ImAxis_X1, NULL, ImPlotAxisFlags_LockMax | ImPlotAxisFlags_LockMin | ImPlotAxisFlags_NoTickLabels);
		ImPlot::DragRect(0, &bprev.X.Min, &bprev.Y.Min, &bprev.X.Max, &bprev.Y.Max,ImVec4(1,0,1,1), ImPlotDragToolFlags_Delayed);

		ImPlot::EndPlot();
	}
	if (ImGui::Button("Copy inside_box() to clipboard")) {
		ImGui::LogToClipboard();
		ImGui::LogText("inside_box(&x, &y, %d, %d, %d, %d)", bprev.X.Min, bprev.Y.Min, bprev.X.Max, bprev.Y.Max);
		ImGui::LogFinish();
	}

	ImGui::End();
}

/*
ImGui::Begin("Extra testing stuff");
if (ImGui::Button("Play ogg")) {
	muhandle = playsfx("sound/sewergangs.ogg", 0, 0);
}

if (ImGui::Button("Play another")) {
	muhandle = playsfx("sound/claimyourprize.mp3", 0, 0);
}

if (ImGui::Button("Play MIDI as SFX")) {
	play_midi_as_sfx("sound/octfalls.mid", "GeneralUser.sf2");
}
ImGui::Text("1 mx: %d", spr[1].mx);
ImGui::Text("1 my: %d", spr[1].my * -1);

if (ImGui::InputText("Lua console input", mbconsole, IM_ARRAYSIZE(mbconsole), ImGuiInputTextFlags_EnterReturnsTrue)) {

	memset(&mbconsole, 0, sizeof(mbconsole));
}

if (ImGui::Button("JSON Dink.dat output")) {
	cereal::JSONOutputArchive archive(std::cout);
	archive(g_dmod.map);
}

if (ImGui::Button("Dink.json save")) {
	std::ofstream os(paths_dmodfile("dink.json"));
	cereal::JSONOutputArchive archive(os);
	archive(g_dmod.map);

}

if (ImGui::Button("Dink.json load")) {
	std::ifstream is(paths_dmodfile("dink.json"));
	cereal::JSONInputArchive archive(is);
	archive(g_dmod.map);
}

if (ImGui::Button("Reload dinklua init")) {
    dinklua_reload_init();
}

if (ImGui::Button("OPN2 emulator")) {
	Mix_OPNMIDI_setEmulator(2);
}

for (int j = 0; j < last_sprite_created; j++) {
	if (spr[j].que != 0)
		ImGui::Text("Sprite %d with cue %d", j, spr[j].que);
	else
		ImGui::Text("Sprite %d with Y co-ord %d", j, spr[j].y);
}
ImGui::End();

/*
ImGui::Begin("Sprite Ranking");

int rankings[MAX_SPRITES_AT_ONCE];
memset(rankings, 0, MAX_SPRITES_AT_ONCE * sizeof(int));
bool already_checked[MAX_SPRITES_AT_ONCE + 1];
memset(already_checked, 0, sizeof(already_checked));

for (int r1 = 0; r1 < last_sprite_created; r1++) {
		rankings[r1] = 0;

		int highest_sprite = 22000; //more than it could ever be

		for (int h1 = 1; h1 <= last_sprite_created; h1++) {
			if (!already_checked[h1] && spr[h1].active) {
				int height;
				if (spr[h1].que != 0)
					height = spr[h1].que;
				else
					height = spr[h1].y;

				if (height < highest_sprite) {
					highest_sprite = height;
					rankings[r1] = h1;
				}
			}
		}

		if (rankings[r1] != 0)
			already_checked[rankings[r1]] = true;
		ImGui::Text("Rank of index %d is sprite %d", r1, rankings[r1]);
	}
	ImGui::Separator();
	//New averaging system
	std::multimap<int, int> myrankings;
	int newrankings[MAX_SPRITES_AT_ONCE];
	memset(newrankings, 0, MAX_SPRITES_AT_ONCE * sizeof(int));
	//get akll sprites and use their Y if no que
	for (int i=0; i <= last_sprite_created; i++) {
			if (spr[i].que != 0) {
				myrankings.insert(std::make_pair(spr[i].que, i));
			}
			else {
				myrankings.insert(std::make_pair(spr[i].y, i));
			}
	}
	int kk = 0;
	for (auto const& entry: myrankings) {
		newrankings[kk] = entry.second;
		ImGui::Text("Rank of index %d is sprite %d", entry.first, entry.second);
		kk++;
	}
	ImGui::Separator();
	for (int i = 1; i <= last_sprite_created; i++) {
		ImGui::Text("Rank of index %d is sprite %d", i, newrankings[i]);
	}

ImGui::End();

/*
ImGui::Begin("Curl");
if (ImGui::Button("init")) {
	CURL *curl;
	CURLcode res;
	curl_global_init(CURL_GLOBAL_ALL);
	curl = curl_easy_init();
	if (curl) {
		struct curl_slist *chunk = NULL;
		static const char *voicefilename = "text.mp3";
		FILE *mpfile;
		curl_easy_setopt(curl, CURLOPT_URL, "https://api.elevenlabs.io/v1/text-to-speech/VR6AewLTigWG4xSOukaG");
		chunk = curl_slist_append(chunk, "accept: audio/mpeg");
		chunk = curl_slist_append(chunk, "xi-api-key: haha nice try");
		chunk = curl_slist_append(chunk, "Content-Type: application/json");
		curl_easy_setopt(curl, CURLOPT_HTTPHEADER, chunk);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDS, "{\"text\": \"Hello hello hello\", \"model_id\": \"eleven_monolingual_v1\", \"voice_settings\": { \"stability\": 0.5, \"similarity_boost\": 0.5}}");
		curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, write_data);
		//res = curl_easy_perform(curl);
		mpfile = fopen(voicefilename, "wb");
  	if(mpfile) {


    curl_easy_setopt(curl, CURLOPT_WRITEDATA, mpfile);


    res = curl_easy_perform(curl);
	 if(res != CURLE_OK)
      fprintf(stderr, "curl_easy_perform() failed: %s\n",
              curl_easy_strerror(res));


    fclose(mpfile);
		curl_easy_cleanup(curl);
		curl_global_cleanup();
	}
}
}
ImGui::End();


ImGui::Begin("Savestate");
if (ImGui::Button("Save to disk")) {
	savestate();
}
if (ImGui::Button("Load from disk")) {
	loadstate();
	game_load_screen(g_dmod.map.loc[*pplayer_map]);
	draw_screen_game();
}
if (ImGui::Button("Reset colours")) {
	std::cout << termcolor::cyan << "Sneed";
	std::cout << termcolor::reset << "World!";
	std::cout << std::endl;
	log_info("TEST");
}
if (ImGui::Button("Fade down")) {
	process_downcycle = 1;
	cycle_clock = game_GetTicks() + 4000;
	//CyclePalette();
}
if (ImGui::Button("Fade up")) {
	process_upcycle = 1;
	//up_cycle();
}
ImGui::Text("Downcycle: %d, Upcycle %d", process_downcycle, process_upcycle);
ImGui::Text("Brightness: %f", g_display->brightness);
ImGui::End();


ImGui::Begin("sp_custom view");
for (int i = 1; i < last_sprite_created; i++) {
	if (spr[i].custom != nullptr) {
		ImGui::Text("Sprite #%d", i);

		for(const auto& value: *spr[i].custom) {
    		ImGui::Text("%s, %d", value.first.c_str(), value.second);
		}
	}
}

ImGui::End();
*/

//Alternate text box drawing - doewsn't work with damage yet
if (dbg.alttext) {
	for (int j = 0; j <= last_sprite_created; j++) {
		//int sprite = rank[j];
		if (spr[j].active && spr[j].brain == 8 && spr[j].damage == -1) {
			//if (spr[j].text_owner != 1000) {
				ImGui::SetNextWindowPos(ImVec2(spr[j].x * (g_display->w / 640.0f), spr[j].y * (g_display->h / 480.0f)));
			//Text colouring
			char crap[200];
			char* cr;
			int color = 0;

		//ye: could probably be optmized
		sprintf(crap, "%s", spr[j].text);
		cr = &crap[0];
		color = 14;
		while (cr[0] == '`') {
			//color code at top
			if (cr[1] == '#') {
				color = 13;
			}
			if (cr[1] == '1') {
				color = 1;
			}
			if (cr[1] == '2') {
				color = 2;
			}
			if (cr[1] == '3') {
				color = 3;
			}
			if (cr[1] == '5') {
				color = 5;
			}
			if (cr[1] == '6') {
				color = 6;
			}
			if (cr[1] == '7')
				color = 7;
			if (cr[1] == '8')
				color = 8;
			if (cr[1] == '9')
				color = 9;
			if (cr[1] == '0')
				color = 10;
			if (cr[1] == '$') {
				color = 14;
			}
			if (cr[1] == '%') {
				color = 15;
			}

			if (dversion >= 108) {
				//support for additional colors
				if (cr[1] == '@')
					color = 12;
				if (cr[1] == '!') {
					color = 11;
				}
			}

			if (cr[1] == '4')
				color = 4;
			cr = &cr[2];
		}

		ImVec4 tcol = ImVec4(font_colors[color].r / 255.0f, font_colors[color].g / 255.0f, font_colors[color].b / 255.0f, 1.0f);

			ImGui::Begin(spr[j].text, NULL, ImGuiWindowFlags_AlwaysAutoResize | ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_NoFocusOnAppearing);
			ImGui::SetWindowFontScale(1.50f);
			ImGui::TextColored(tcol, cr);
			ImGui::End();
		}
	}
}

if (shaderwindow) {
ImGui::Begin("Shaders", &shaderwindow);
ImGui::Text("Only available under the OpenGL backend");
ImGui::SeparatorText("Colour Adjustment");
ImGui::Checkbox("Enable", &dbg.shader_colour);
ImGui::Checkbox("Autocycle", &dbg.shader_colcycle);
ImGui::SliderFloat("R", &dbg.shader_r, 0.0f, 1.0f);
ImGui::SliderFloat("G", &dbg.shader_g, 0.0f, 1.0f);
ImGui::SliderFloat("B", &dbg.shader_b, 0.0f, 1.0f);
ImGui::Spacing();
if (ImGui::Button("Load Shader")) {
//	Uint32 v, f, *p;
//	v = GPU_LoadShader(GPU_VERTEX_SHADER, paths_pkgdatafile("common.vert"));
//	f = GPU_LoadShader(GPU_FRAGMENT_SHADER, paths_pkgdatafile("common.frag"));
//	*p = GPU_LinkShaders(v, f);
//	GPU_ShaderBlock block = GPU_LoadShaderBlock(*p, "gpu_Vertex", "gpu_TexCoord", "gpu_Color", "gpu_ModelViewProjectionMatrix");
//	GPU_ActivateShaderProgram(*p, &block);
}
ImGui::End();
}


//Stuff that runs each frame that isn't an imgui window
//To make it pause when console is shown and relevant setting is on
if (console_active == 0 && debug_conpause) {
	debug_paused = 0;
}
//Stolen from freedinkedit, toggled in display options, requested by drone1400. Runs really slow in GL mode
if (drawhard) {
    int x1, y1;
	for (x1 = 0; x1 < 600; x1++)
		for (y1 = 0; y1 < 400; y1++) {
			if (screen_hitmap[x1][y1] == 1) {
				{
					SDL_Rect GFX_box_crap;
					GFX_box_crap.x = x1 + playl;
					GFX_box_crap.y = y1;
					GFX_box_crap.w = 1;
					GFX_box_crap.h = 1;
					IOGFX_backbuffer->fillRect(&GFX_box_crap, GFX_ref_pal[1].r,
											GFX_ref_pal[1].g,
											GFX_ref_pal[1].b);
				}
			}

			if (screen_hitmap[x1][y1] == 2) {
				{
					SDL_Rect GFX_box_crap;
					GFX_box_crap.x = x1 + playl;
					GFX_box_crap.y = y1;
					GFX_box_crap.w = 1;
					GFX_box_crap.h = 1;
					IOGFX_backbuffer->fillRect(
							&GFX_box_crap, GFX_ref_pal[128].r,
							GFX_ref_pal[128].g, GFX_ref_pal[128].b);
				}
			}

			if (screen_hitmap[x1][y1] == 3) {
				{
					SDL_Rect GFX_box_crap;
					GFX_box_crap.x = x1 + playl;
					GFX_box_crap.y = y1;
					GFX_box_crap.w = 1;
					GFX_box_crap.h = 1;
					IOGFX_backbuffer->fillRect(&GFX_box_crap, GFX_ref_pal[45].r,
											GFX_ref_pal[45].g,
											GFX_ref_pal[45].b);
				}
			}

			if (screen_hitmap[x1][y1] > 100) {

				if (cur_ed_screen.sprite[(screen_hitmap[x1][y1]) - 100]
							.is_warp == 1) {
					{
						SDL_Rect GFX_box_crap;
						GFX_box_crap.x = x1 + playl;
						GFX_box_crap.y = y1;
						GFX_box_crap.w = 1;
						GFX_box_crap.h = 1;
						IOGFX_backbuffer->fillRect(
								&GFX_box_crap, GFX_ref_pal[20].r,
								GFX_ref_pal[20].g, GFX_ref_pal[20].b);
					}
				} else {
					{
						SDL_Rect GFX_box_crap;
						GFX_box_crap.x = x1 + playl;
						GFX_box_crap.y = y1;
						GFX_box_crap.w = 1;
						GFX_box_crap.h = 1;
						IOGFX_backbuffer->fillRect(
								&GFX_box_crap, GFX_ref_pal[23].r,
								GFX_ref_pal[23].g, GFX_ref_pal[23].b);
					}
				}
			}
		}

}
//Draws a box around the sprite bounds so you can see which one you're editing. Toggled in the live sprite editor.
if (spr[spriedit].active == 1 && (sprinfo || spreditors)) {
if (spoutline) {
	rect box;
	rect_copy(&box, &k[getpic(spriedit)].box);
	//Now uses imgui windows rather than SDL itself
	ImGui::SetNextWindowPos(ImVec2((box.left + spr[spriedit].x - k[getpic(spriedit)].xoffset) * (g_display->w / 640), (spr[spriedit].y - k[getpic(spriedit)].yoffset) * (g_display->h / 480)));
	ImGui::SetNextWindowSize(ImVec2(box.right * (g_display->w / 640.0f), box.bottom * (g_display->h / 480.0f)));
	ImGui::PushStyleColor(ImGuiCol_Border, ImVec4(1.0f, 0.2f, 0.2f, 1.0f));
	ImGui::PushStyleColor(ImGuiCol_WindowBg, ImVec4(0.0f, 0.2f, 0.2f, 0.2f));
	ImGui::Begin("##spriteoutline", NULL, ImGuiWindowFlags_NoTitleBar | ImGuiWindowFlags_NoFocusOnAppearing | ImGuiWindowFlags_NoResize);
	ImGui::PopStyleColor(2);
	ImGui::End();
	//SDL_Rect r = {box.left + spr[spriedit].x - k[getpic(spriedit)].xoffset, box.top + spr[spriedit].y - k[getpic(spriedit)].yoffset, box.right - box.left,
	//								box.bottom - box.top};
	//					IOGFX_backbuffer->fillRect(&r, 255, 100, 0);
}
//hardbox preview for the sprite editor
if (hboxprev) {
	rect hbox;
	rect_copy(&hbox, &k[getpic(spriedit)].hardbox);
	SDL_Rect h = {hbox.left + spr[spriedit].x, hbox.top + spr[spriedit].y, hbox.right - hbox.left,
									hbox.bottom - hbox.top};
	IOGFX_backbuffer->fillRect(&h, 0, 100, 255);
}
//Depth dot
if (ddotprev) {
	rect ddot;
	ddot.left = spr[spriedit].x - 2;
	ddot.right = ddot.left + 3;
	ddot.top = spr[spriedit].y - 2;
	ddot.bottom = ddot.top + 3;
	SDL_Rect i = {ddot.left, ddot.top, 2, 2};
	IOGFX_backbuffer->fillRect(&i, 255,10,0);
}
}
//turns on overlay thingy
if (console_active && !dbg.dswitch && !logwindow)
dbg.dswitch = 1;

}
}
