/**
 * Sound effects (not music) management

 * Copyright (C) 1997, 1998, 1999, 2002, 2003  Seth A. Robinson
 * Copyright (C) 2003  Shawn Betts
 * Copyright (C) 2005, 2007, 2008, 2009, 2014, 2015  Sylvain Beucler
 * Copyright (C) 2022, 2023 CRTS, Yeoldetoast

 * This file is part of Yeoldedink

 * Yeoldedink is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.

 *  Yeoldedink is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <string.h> /* memset, memcpy */
#include <errno.h>

#include "SDL.h"
#ifdef HAVE_MIXER_EXT
#include <SDL2/SDL_mixer_ext.h>
#else
#include "SDL_mixer.h"
#endif

#include "soloud/soloud.h"
#include "soloud/soloud_wav.h"
#include "soloud/soloud_freeverbfilter.h"
#include "soloud/soloud_bassboostfilter.h"
#include "soloud/soloud_speech.h"
#include "soloud/soloud_midi.h"
#include "random.hpp"

#include "live_sprites_manager.h"
#include "debug_imgui.h"
#include "game_engine.h"
#include "live_screen.h"
#include "io_util.h"
#include "paths.h"
#include "log.h"
#include "math.h"
#include "sfx.h"
#include "gfx_sprites.h"
#include "log.h"

//for pitch randomisation
using Random = effolkronium::random_static;

SoLoud::Soloud gSoloud;
SoLoud::Wav gWave;
SoLoud::Bus gBus;
SoLoud::FreeverbFilter gVerb;
SoLoud::BassboostFilter gBassBoost;
SoLoud::Speech speech;
SoLoud::Midi mymidi;
SoLoud::SoundFont mysoundfont;

/* Sound - BGM */
int sound_on = 1;

/* Channel metadata */
int buf_samples = 1024;
int audio_samplerate;
int adlemu, opnemu;
char soundfont[200];
//3d sound parameters
float atten = 0.8f;
float dist_min = 20.0f;
float dist_max = 610.0f;
int dist_model = 2;
int distmult = 100;

/* Sound metadata */
#define MAX_SOUNDS 100

//register our soloud sound slots to load into
SoLoud::Wav gSFX[MAX_SOUNDS];
//create groups so we can get repeating and survive
SoLoud::handle gh = gSoloud.createVoiceGroup();
SoLoud::handle gs = gSoloud.createVoiceGroup();
//Main queue
SoLoud::Queue gQueue;

/* Hardware soundcard information */
static int hw_freq, hw_channels;
static Uint16 hw_format;

/**
 * Display a SDL audio-format in human-readable form
 **/
static const char* format2string(Uint16 format) {
	char* format_str = "Unknown";
	switch (format) {
	case AUDIO_U8:
		format_str = "U8";
		break;
	case AUDIO_S8:
		format_str = "S8";
		break;
	case AUDIO_U16LSB:
		format_str = "U16LSB";
		break;
	case AUDIO_S16LSB:
		format_str = "S16LSB";
		break;
	case AUDIO_U16MSB:
		format_str = "U16MSB";
		break;
	case AUDIO_S16MSB:
		format_str = "S16MSB";
		break;
	case AUDIO_S32LSB:
		format_str = "S32LSB";
		break;
	case AUDIO_S32MSB:
		format_str = "S32MSB";
		break;
	case AUDIO_F32LSB:
		format_str = "F32LSB";
		break;
	case AUDIO_F32MSB:
		format_str = "F32MSB";
		break;
	}
	return format_str;
}
/**
 * Load sounds from the standard paths. Sound is converted to the
 * current hardware (soundcard) format so that sound rate can be
 * altered with minimal overhead when playing the sound.
 */
int CreateBufferFromWaveFile(char* filename, int index) {
	/* Open the wave file */
	char path[150];

	if (index >= MAX_SOUNDS) {
		log_error("SCRIPTING ERROR: sound index %d is too big.", index);
		return 0;
	}
	sprintf(path, "sound/%s", filename);
	//Loading first from main data, then d-mod. Seems to work
	if (sound_on) {
	gSFX[index].load(paths_fallbackfile(path));
	gSFX[index].load(paths_dmodfile(path));
	//Setting this improves the blast thingy that occurs upon changing screens, but may annoy some
	gSFX[index].set3dDistanceDelay(dbg.audiodelay);
	}

	return 1;
}

/**
 * Is the specified sound currently playing?
 *
 * Only used in pig_brain(), could be removed maybe.
 */
int playing(int sound) {
	return (gSoloud.isValidVoiceHandle(sound));
}

/**
 * Kill repeating sounds except the ones that survive
 */
void kill_repeat_sounds(void) {
	if (!sound_on)
		return;

	log_info("🔇 Killing non-surviving repeating sounds");
	//Get rid of all those not set to survive
	gSoloud.fadeVolume(gh, 0, 0.3f);
	gSoloud.setVolume(gs, 1.0f);
	gSoloud.schedulePause(gh, 1);
	gSoloud.setPause(gs, false);
}

/**
 * Kill all repeating sounds, even the ones that survive (used from
 * DinkC's restart_game() and load_game())
 */
void kill_repeat_sounds_all(void) {
	if (!sound_on)
		return;

	gSoloud.stop(gh);
}

/**
 * Kill one sound
 */
void kill_this_sound(int channel) {
	gSoloud.stop(channel);
}

//yeolde: Actually just halt all SFX
void halt_all_sounds() {
	gSoloud.stopAll();
}

/**
 * Called by update_frame()
 *
 * If sound is active, refreshed pan&vol for 3D effect.
 *
 * If repeating and sprite.sound==0 and sprite.owner!=0 -> stop sound
 * If sprite.active==0 and sprite.owner!=0 -> stop sound
 */
void update_sound(void) {
	int i;

	if (!sound_on)
		return;

	for (i = 1; i < MAX_SPRITES_AT_ONCE; i++) {
			if (spr[i].active && gSoloud.isVoiceGroup(spr[i].sounds)) {
			//This sprite is alive. Let's update its positioning.
			//int z = k[getpic(i)].box.bottom * k[getpic(i)].box.right;
			gSoloud.set3dSourceParameters(spr[i].sounds, spr[i].x, spr[i].y, 0,
				spr[i].mx, spr[i].my, 0);
			}
		}
	//TODO: delay this until the screen transition is complete somehow
	//int z = k[getpic(1)].box.bottom * k[getpic(1)].box.right;
	//Set the player's listening pos
	gSoloud.set3dListenerPosition(spr[1].x, spr[1].y, 0);
	gSoloud.set3dListenerVelocity(spr[1].mx, spr[1].my * -1, 0);
	//This doesn't work properly yet
	if (dbg.debug_audiopos)
	gSoloud.set3dListenerUp(spr[1].mx, spr[1].my * -1, 0);
	else
	gSoloud.set3dListenerUp(0, -1, 0);

	gSoloud.update3dAudio();
}

static int SoundPlayEffectChannel(int sound, int min, int plus, int sound3d,
								/*bool*/ int repeat, int explicit_channel);

/**
 * Play a sound previously loaded to memory (in registered_sounds)
 * - sound: sound index
 * - min: frequency (Hz)
 * - plus: max random frequency, to add to min
 * - sound3d: if != 0, sprite number whose location will be used for
 *   pseudo-3d effects (volume, panning)
 * - repeat: is sound looping?
 **/
int SoundPlayEffect(int sound, int min, int plus, int sound3d,
					/*bool*/ int repeat) {
	return SoundPlayEffectChannel(sound, min, plus, sound3d, repeat, -1);
}
/**
 * SoundPlayEffect_Channel_ allows to specify an explicit audio
 * channel, which in turns allows the editor to only use one channel
 * for everything (when you move the mouse with the keyboard, you'll
 * hear a series of close 'ticks', but they won't overlap each
 * others). The rest of the time, the game will just pass '-1' for the
 * channel, so the first available channel (among NUM_CHANNELS useable
 * simultaneously) will be selected.
 */
static int SoundPlayEffectChannel(int sound, int min, int plus, int sound3d,
								/*bool*/ int repeat, int explicit_channel) {
	log_enter("SoundPlayEffectChannel: sound: %d, min: %d, plus: %d, "
			"sound3d: %d, repeat: %d, explicit_channel: %d", sound, min,
			plus, sound3d, repeat, explicit_channel);

	//Safety check. Don't know if it's needed due to Soloud's virtual voices
	//if (gSoloud.getActiveVoiceCount() == gSoloud.getMaxActiveVoiceCount()) {
	//	log_error("Out of voices for playsound()");
	//	return -1;
	//}


	//Our audio handle number to return so they can do stuff with it
	int x;
	//Set our listener before playing a sound
	gSoloud.set3dListenerPosition(spr[1].x, spr[1].y, spr[1].size / 100.0f);
	/* Sample rate / frequency */
	{
		int play_freq;

		if (plus == 0)
			play_freq = min;
		else {
			play_freq = Random::get(min, min + plus -1);
			//log_info("Play freq is min %d with %d equalling %d", min, plus, play_freq);
		}

		if (sound3d > 0) {
			//Sound3d is a sprite number that we'll use for positioning
			//log_error("Sound3d value is %d", sound3d);
			//log_error("3d sprite is %d and %d", spr[sound3d].x, abs(spr[sound3d].y - 400));
			//log_error("sprite1 x is %d, and y is %d", spr[1].x, abs(spr[1].y - 400));
			x = gSoloud.play3d(gSFX[sound], spr[sound3d].x, spr[sound3d].y, 0, spr[sound3d].mx, spr[sound3d].my, 0, 1.0f, true);
			//Make it attenuate as we walk away from it. Higher values increase the rolloff
			gSoloud.set3dSourceMinMaxDistance(x, dist_min, dist_max);
			gSoloud.set3dSourceAttenuation(x, dist_model, atten);
			//use an extra field in the sprite to store the handle(s)
			if (gSoloud.isVoiceGroup(spr[sound3d].sounds)) {
				gSoloud.addVoiceToGroup(spr[sound3d].sounds, x);
			}
			else {
				//Create the group for the sprite if it doesn't already exist
				spr[sound3d].sounds = gSoloud.createVoiceGroup();
				gSoloud.addVoiceToGroup(spr[sound3d].sounds, x);
			}
		} else {
			//A normal, non-positional sound not attached to a sprite we'll play clocked to avoid bunching up
			x = gSoloud.playClocked(time_start, gSFX[sound]);

			//x = gSoloud.play(gSFX[sound], 1.0f, 0.0f, true);
		}
		//Kill the sound if it becomes inaudible
		gSoloud.setInaudibleBehavior(x, false, true);
		//Set the speed in hertz, or else use defaults if set to zero
		if (min != 0)
		gSoloud.setSamplerate(x, play_freq);

		if (repeat == 1) {
		gSoloud.setLooping(x, true);
		//Keep the sound ticking even if it's inaudible
		gSoloud.setInaudibleBehavior(x, true, false);

			if (!sound3d)
				gSoloud.addVoiceToGroup(gh, x);
		}
		//Actually play it out loud if not clocked
		gSoloud.setPause(x, false);

	}
	log_exit("SoundPlayEffectChannel: %d", x);
	return x;
}

/**
 * SoundStopEffect
 *
 * Stops the sound effect specified.
 * Returns TRUE if succeeded.
 */
int SoundStopEffect(int sound) {

	if (sound >= MAX_SOUNDS) {
		log_error("🎸 Attempting to get stop sound %d (> MAX_SOUNDS=%d)", sound,
				MAX_SOUNDS);
		return 0;
	} else{
		gSoloud.stopAudioSource(gSFX[sound]);
		return 1;
	}

}

/**
 * InitSound
 *
 * Sets up the DirectSound object and loads all sounds into secondary
 * DirectSound buffers.  Returns -1 on error, or 0 if successful
 */
void sfx_init() {
	if (!sound_on) {
		//Init soloud with no sound output
		gSoloud.init(SoLoud::Soloud::FLAGS::LEFT_HANDED_3D, SoLoud::Soloud::BACKENDS::NULLDRIVER, SoLoud::Soloud::AUTO, SoLoud::Soloud::AUTO, SoLoud::Soloud::AUTO);
		log_info("🔇 SoLoud initted with no audio");
		return;
	}

	sound_on = 0;

	log_info("📻 Initting sound");

	#ifdef _WIN32
	SDL_setenv("SDL_AUDIODRIVER", "wasapi", 1);
	#endif
	if (SDL_Init(SDL_INIT_AUDIO) == -1) {
		log_error("📻 SDL_Init(SDL_INIT_AUDIO): %s", SDL_GetError());
		return;
	}

	/* Work-around to disable fluidsynth and fallback to TiMidity++: */
	/* TODO: allow user to set it at run-time */
	/* SDL_setenv("SDL_SOUNDFONTS="); */
	/* SDL_setenv("SDL_FORCE_SOUNDFONTS=1"); */
	//Yeolde: set in ini file
	if (strcmp(soundfont, "0"))
	Mix_SetSoundFonts(paths_pkgdatafile(soundfont));
	/* To specify an alternate timidity.cfg */
	/* Cf. SDL2_mixer/timidity/config.h for defaults */
	/* SDL_setenv("TIMIDITY_CFG", "somewhere/timidity.cfg", 0); */

	/* MIX_DEFAULT_FREQUENCY is ~22kHz are considered a good default,
44kHz is considered too CPU-intensive on older computers */
//yeolde: Thankfully I don't care about them.
	/* MIX_DEFAULT_FORMAT is 16bit adapted to current architecture
(little/big endian) */
	/* MIX_DEFAULT_CHANNELS is 2 => stereo, allowing panning effects */
	/* 1024 (chunk on which effects are applied) seems a good default,
4096 is considered too big for SFX */

	//yeolde: even with SoLoud we still need to init Mixer for bgm
	if (Mix_OpenAudio(audio_samplerate, MIX_DEFAULT_FORMAT,
					MIX_DEFAULT_CHANNELS, buf_samples) == -1) {
		log_error("Mix_OpenAudio: %s", Mix_GetError());
		return;
	}

	/* Done with initialization */
	/* Avoid calling SDL_PauseAudio when using SDL_Mixer */
	/* SDL_PauseAudio(0); */

	/* Dump audio info */
	{
		log_info("📻 Audio driver: %s", SDL_GetCurrentAudioDriver());

		int numtimesopened;
		numtimesopened = Mix_QuerySpec(&hw_freq, &hw_format, &hw_channels);
			log_info("📻 Audio hardware info: "
					"frequency=%dHz\tformat=%s\tchannels=%d\topened=%d times",
					hw_freq, format2string(hw_format), hw_channels,
					numtimesopened);
	}

	sound_on = 1;
	//Set our saved parameters
	Mix_VolumeMusic(dbg.musvol);
	#ifdef HAVE_MIXER_EXT
	Mix_SetMidiPlayer(midiplayer);
	Mix_OPNMIDI_setEmulator(opnemu);
	Mix_ADLMIDI_setEmulator(adlemu);
	#endif
	//init soloud
	//Backend defined in soloud.h - could be made user-configurable
	#ifdef WITH_MINIAUDIO
	gSoloud.init(SoLoud::Soloud::FLAGS::LEFT_HANDED_3D, SoLoud::Soloud::BACKENDS::MINIAUDIO, audio_samplerate, SoLoud::Soloud::AUTO, 2);
	#elif defined WITH_SDL2_STATIC
	gSoloud.init(SoLoud::Soloud::FLAGS::LEFT_HANDED_3D, SoLoud::Soloud::BACKENDS::SDL2, audio_samplerate, buf_samples, 2);
	#endif
	log_info("📻 SoLoud initted at %dHz with buffer of %d using %s with %d channels", gSoloud.getBackendSamplerate(), gSoloud.getBackendBufferSize(), gSoloud.getBackendString(), gSoloud.getBackendChannels());
	//let's give them lots of voices
	gSoloud.setMaxActiveVoiceCount(NUM_CHANNELS);
	//For our waveform viewer
	gSoloud.setVisualizationEnable(true);
	//Set our 3d vector, sound speed, and clipper
	gSoloud.set3dListenerUp(1, -1, 0);
	gSoloud.setPostClipScaler(0.9);
	//Set to 10x the speed of sound, assuming that 1 in-game unit is 100m or so
	gSoloud.set3dSoundSpeed(3430);
	gVerb.setParams(0, 2, 5, 5);
	//gSoloud.setGlobalFilter(0, &gVerb);
	speech.setParams(2000, 6, 2, KW_TRIANGLE);
	//Set our loaded sfx volume
	gSoloud.setGlobalVolume(dbg.sfxvol);
	//Set our speech bus bass boost
	gBassBoost.setParams(0.8f);
	gBus.setFilter(0, &gBassBoost);
}

/**
 * Undoes everything that was done in a sfx_init call
 */
void sfx_quit(void) {
	if (!sound_on)
		return;

	/* Stops all SFX channels */
	//gSoloud.stopAll();
	gSoloud.deinit();
	Mix_Quit();
	Mix_CloseAudio();
	SDL_QuitSubSystem(SDL_INIT_AUDIO);
}

/**
 * Print SFX memory usage
 */
void sfx_log_meminfo() {
	int sum = sizeof(gSFX);
	log_debug("Sounds   = %8d", sum);
}


/**
 * Set volume; dx_volume is [-10000;10000] in hundredth of dB
 */
//TODO: figure out discrepancies vs RTDink. Sounds much louder on there with volume increases
static int SetVolume(int channel, int dx_volume) {
	// SFX
	/* See doc/sound.txt for details */
	if (!gSoloud.isValidVoiceHandle(channel))
		log_warn("👹 Attempted to set the volume of invalid voice handle!");

	gSoloud.setVolume(channel, dx_volume / 10000.0f);
	//return Mix_Volume(channel,
	//				MIX_MAX_VOLUME * pow(10, ((double)dx_volume / 100) / 20));
	return 1;
}

/** DinkC procedures **/
/* BIG FAT WARNING: in DinkC, soundbank is channel+1
(a.k.a. non-zero), and 0 means failure. */
int playsound(int sound, int min, int plus, int sound3d, int repeat) {
	log_enter("playsound: sound: %d, min: %d, plus: %d, sound3d: %d, repeat: %d",
			sound, min, plus, sound3d, repeat);
	int channel = SoundPlayEffect(sound, min, plus, sound3d, repeat);
	log_debug("🎵 Sound played on handle: %d", channel);

	log_exit("playsound: %d", channel);
	return channel;
}

void sound_set_kill(int soundbank) {
	gSoloud.stop(soundbank);
}

void sound_set_survive(int soundbank, int survive) {
	if (survive) {
	gSoloud.setProtectVoice(soundbank, true);
	gSoloud.addVoiceToGroup(gs, soundbank);
	}
	else {
		gSoloud.setProtectVoice(soundbank, false);
		//No way to remove voices from group, I don't think
	}
}

void sound_set_vol(int soundbank, int volume) {
	SetVolume(soundbank, volume);
}


//For Lua only
void lua_set_vol(int soundbank, int volume) {
	if (!gSoloud.isValidVoiceHandle(soundbank) || volume == 0)
		log_warn("👹 Attempted to set the volume of invalid voice handle!");
	gSoloud.setVolume(soundbank, volume / 100.0f);
}

int sound_get_vol(int soundbank) {
	return gSoloud.getVolume(soundbank) * 100;
}

void sound_fade_vol(int soundbank, int vol, int time) {
	gSoloud.fadeVolume(soundbank, vol / 100.0f, time);
}

bool sound_get_survive(int soundbank) {
	//Only using voice protection for survived sounds I think
	return gSoloud.getProtectVoice(soundbank);
}

void sound_set_pan(int soundbank, int pan) {
	gSoloud.setPan(soundbank, pan / 100.0f);
}

void sound_fade_pan(int soundbank, int pan, int time) {
	gSoloud.fadePan(soundbank, pan / 100.0f, time);
}

int sound_get_pan(int soundbank) {
	return (int)gSoloud.getPan(soundbank * 100);
}

void sound_set_hz(int soundbank, int hz) {
	gSoloud.setSamplerate(soundbank, hz);
}

int sound_get_hz(int soundbank) {
	return (int)gSoloud.getSamplerate(soundbank);
}

int sound_get_speed(int soundbank) {
	return (int)gSoloud.getRelativePlaySpeed(soundbank) * 100;
}

void sound_set_speed(int soundbank, int speed) {
	//Sound speed default is 100
	if (speed > 0)
		gSoloud.setRelativePlaySpeed(soundbank, speed / 100.0f);
}

void sound_fade_speed(int soundbank, int speed, int time) {
	if (speed > 0)
	gSoloud.fadeRelativePlaySpeed(soundbank, speed / 100.0f, time);
}

void sound_set_pause(int soundbank, bool paused) {
	gSoloud.setPause(soundbank, paused);
}

bool sound_get_pause(int soundbank) {
	return gSoloud.getPause(soundbank);
}

void sound_schedule_pause(int soundbank, int time) {
	gSoloud.schedulePause(soundbank, time);
}

void sound_schedule_stop(int soundbank, int time) {
	gSoloud.scheduleStop(soundbank, time);
}

void sound_set_loop_point(int soundbank, int looppoint) {
	gSoloud.setLoopPoint(soundbank, looppoint / 1.0f);
}

int sound_get_loop_point(int soundbank) {
	return (int)gSoloud.getLoopPoint(soundbank);
}

int sound_get_loop_count(int soundbank) {
	return gSoloud.getLoopCount(soundbank);
}

int sound_get_pos(int soundbank) {
	return gSoloud.getStreamPosition(soundbank);
}

void sound_oscillate_vol(int soundbank, int from, int to, int time) {
	gSoloud.oscillateVolume(soundbank, from / 100.0f, to / 100.0f, time);
}

void sound_oscillate_pan(int soundbank, int from, int to, int time) {
	gSoloud.oscillatePan(soundbank, from / 100.0f, to / 100.0f, time);
}

void sound_oscillate_speed(int soundbank, int from, int to, int time) {
	gSoloud.oscillateRelativePlaySpeed(soundbank, from / 100.0f, to / 100.0f, time);
}

void sound_set_looping(int soundbank, bool repeat) {
	gSoloud.setLooping(soundbank, repeat);
}

bool sound_get_looping(int soundbank) {
	return gSoloud.getLooping(soundbank);
}

//yeolde: use soloud to one-shot play a wav/ogg/flac
int playsfx(char* filename, int speed, int pan) {
	//gWave.load(paths_dmodfile(filename));
	gWave.load(filename);
	int x = gSoloud.play(gWave);
	if (speed != 0) {
		gSoloud.setRelativePlaySpeed(x, speed / 100.0f);
	}
	if (pan != 0) {
		gSoloud.setPan(x, pan / 100.0f);
	}
	return x;
	}

void play_midi_as_sfx(char* filename, char* sf2) {
	//todo: check files exist first
	mysoundfont.load(paths_pkgdatafile(sf2));
	mymidi.load(paths_dmodfile(filename), mysoundfont);
	int x = gSoloud.play(mymidi);
}
